--[[
Puzzle Fighter v1.0.0 - by Xinhuan (Blackrock Alliance <No MAAM>, Gnome Mage)
(7 Aug 2008)
=============================================================================

Known Issues:
- Due to latency, it is possible for 2 players to lose simultaeneously
  (i.e both sides "Game Over" in the same second before the other side learns
  about it due to ping).

Tricks:
- You can hold down CTRL and SHIFT and click on START to play a game against
  yourself. You cannot challenge yourself normally.

Impossible to Implement:
- Auto-pausing the game on zoning (loading screen) is not possible. This is
  because SendAddonMessage() does not work during a PLAYER_LEAVING_WORLD
  event.

Disclaimer:
- This game is NOT secure! Anyone with coding knowledge can cheat easily by
  using macros or modifying the addon code. Do not gamble money or items,
  real or imaginary.

Notes:
- This addon is first written in late 2006. The addon was nearly completed
  by the end of the year, but patch 2.0 broke it. The addon was not fixed.
  WowInterface launched a Crazy Mods contest at
  http://www.wowinterface.com/page-gonecrazy.html in July 2008 and
  hence this addon is revived.

  As a consequence of this, the coding style of the addon uses some 800+
  globals with a huge XML file which is neither efficient nor good programming
  practice at all. In the interest of the contest, the addon is not rewritten
  due to the 8 August 2008 deadline, but just fixed to work.
 

History
=======
(10 Aug 2006)
v0.1	- Basic frames/buttons/interface created.

(11 Aug 2006)
v0.2	- Logic for moving left/right/down and rotate added.
	- Added timing and animation for gem movement.

(12 Aug 2006)
v0.3	- Added code for small gems to coalesce into larger ones.

(13 Aug 2006)
v0.4	- Added 4 crash gems and the code for destroying the gems.

(14 Aug 2006)
v0.5	- Added code to collapse the field with animation.
	- And repeatedly coalesce/crash again if necessary.
	- Added a "next piece" preview.
	- Added diamond crash gem.
	- (The "single player" component is basically complete.)

(17 Aug 2006)
v0.6	- Started on the multiplayer aspect of the game.
	- Added a second player field and modified code to work on both 1P and 2P fields.
	- Implemented basic dialogs for challenging, accepting and declining a game.
	- Implemented basic communication between 2 players to watch each other's game.
	- Implemented code to block out spam (addon communication) generated by the addon.

(18 Aug 2006)
v0.6.1	- Fixed bug with pausing the game on hiding the frame (pause flag not set).
	- Fixed bug where the game sent multiple gems to be spawned at the start of a game.
	- Added checking code against malicious PMs with invalid arguments.

(19 Aug 2006)
v0.7.0	- Rewrote the handling of chat communication to be independant of chat frames.
	- Rewrote portions of code to allow initalisation of game settings in stages (for future use).
	- Removed a few redundant variables and shifted some code into functions.
	- Implemented the same string of gems to spawn for both players in a multiplayer game.
	- Added code to implement multiplayer Game Over.
	- Added informational splash display of "Chain Combos" and "Game Over".
	- Added code to prevent simple cheating via sending PMs with "PF_" prefixes (chat editbox).
	- Added sound effects to the game.

(20 Aug 2006)
v0.7.1	- Fixed the bugged display that says "0-Chain" on the opponent's playing field.
	- Fixed the swapped sound effects on Winning/Losing a game.

(25 Aug 2006)
v0.8.0	- Rewrote the button retrieval code to be more efficient and clear. Instead of using
	  getglobal() on concantenated strings to get a button, it now uses an array of button
	  references directly.
	- Most objects (and code) are also renamed with _P1 and _P2 tags for code consistency.
	- Simplified some minor code with math library functions.
	- Implemented Counter Gems (Ken pattern) that drop in the opponent's field after an attack.
	- Implemented Counter Gems to work with Diamond Crash Gems.
	- Implemented Counter Gems to countdown per turn.
	- Attacks can neutralise incoming attacks at half the rate. Remainder will attack the enemy.
	- Counter Gems that are next to a crash gem that is crashing will also crash.
	- Increased the rate (chance) of crash gem spawns by 50%.

(26 Aug 2006)
v0.8.1	- Fixed a graphical display bug. When an existing power gem in P2's field extends to the
	  left or to the bottom to form a larger power gem, it would previously "disappear" from
	  P2's field and appear in P1's field instead (and cause what looks like overlap issues).
	- Counter Gems that are next to a normal gem that is crashing will also crash.
	- Fixed a bug where counter gems chose to fall in columns 123465 instead of 123564 in that
	  order.
	- Added display counter for each player that indicates how many counter gems will fall.
	- Added a "Tech Bonus" splash display if a diamond breaks on nothing (row 1), or if the
	  player clears the field totally.
	- The very first pair of Gem Spawns (at the start of a game) will no longer be a pair
	  that destroys itself. It will always be 2 normal gems. This is to prevent first turn
	  tech bonus of clearing the whole field.
	- Added sound effect (quest accept) for accepting a challege.

(27 Aug 2006)
v0.9.0	- Cleaned up initialisation code.
	- Implemented all 11 Counter Gem pattern characters (was just Ken previously).
	- Implemented interface to select your character prior to the game starting after accepting
	  a challenge.
	- Implemented a miniature gem pattern of the 2 player heroes.

(31 Aug 2006)
v0.9.1	- Key Bindings has been added for Rotate, Move Left/Right/Down and Drop Down. No graphical
	  button or option exists yet for the Drop Down function though.
	- XML and associated code modified to allow localization in the future. The LUA code is not
	  localized yet though.
	- Fixed another display bug that may occur while on low FPS. Occurs when a gem is falling
	  (after a crash), then merges with another gem on the left (bottom). If the animation of
	  the falling gem isn't over before the coalesce code begins to execute, it causes the
	  display of the resultant merged gem to be displaced to the right (top) visually by one
	  column (row).
	- Reset button now properly resets the Start button (enables it back). If it is disabled
	  during character selection but the user resetted, it previously did not reenable the
	  Start button.

(1 Sep 2006)
v0.9.2	- Rewrote the communication model on the receiving end to use a message buffer. This is to
	  fix synchronization problems when one player is playing too fast for the other player to
	  "process" (a result of implementing key bindings and "drop down" in v0.9.1).
	- Implemented a better Pause/Unpause that is less abusable.
	- Implemented multiplayer pause.
	- Localization of Puzzle Fighter is now possible via localization.lua
	- v0.9.2 is still playable with older v0.9.x, but players using the older versions will NOT
	  have multiplayer pause. Older versions may also get de-sync with the opponent playing
	  field if the opponent is playing too fast.

(10 Sep 2006)
v0.9.3	- Added Options Frame.
	- Added option to lock the playing frame from accidental movement.
	- Added option to configure the "Drop Down" button.
	- Added option to scale the main playing frame between 50% to 100%.
	- Added option to set the transparency of the main playing frame between 25% to 100%.
	- Added option to turn off sound effects.

(26 Sep 2006)
v0.9.9	- Added "Random" character selection.

(26 Jul 2008)
v0.9.91	- Fix API that broke since patch 2.0 two years ago with Lua 5.1 upgrade.
	- Convert PM whisper communications to use SendAddonMessage().
	- Perform some minor optimizations.
	- Throtle OnUpdates so that fast machines do not have an advantage.
	- Spawn the gem one square higher.
	- Remove unnecessary semicolons.
	- Fix bug with playing against yourself where the selected hero doesn't match.
	- Fix animation intervals.

(29 Jul 2008)
v0.9.92	- Fix a situation where there can be up to 86 gems on screen for one player
	  but only 84 are coded.
	- Allow challenges to be sent/received after a Game Over before resetting.
	- Change the Down Button to DROP function to be on by default.

(5 Aug 2008)
v0.9.93	- Tweak attack formula slightly to round up instead of round down.
	- Add help button and help text for new players.
	- Perform some minor optimizations.
	- Removed unneccessary brackets in code.
	- Remove some globals.

(7 Aug 2008)
v1.0.0	- Public launch version.


*** Possible Post-Release Features
- Add version check?
- Allow 3rd party to watch a game in progress (requires use of some sync channel)
  This could really be too much of a hassle though.
- Perhaps implement options for skins, base code is already programmed in a way that allows it.
]]


--------- Game Data and Constants -----------

-- Some strings for the key bindings
BINDING_HEADER_PUZZLE_FIGHTER	= PF_PUZZLE_FIGHTER_VERSION_TEXT
BINDING_NAME_PF_ROTATE		= PF_ROTATE_TEXT
BINDING_NAME_PF_LEFT		= PF_MOVE_LEFT_TEXT
BINDING_NAME_PF_RIGHT		= PF_MOVE_RIGHT_TEXT
BINDING_NAME_PF_DOWN		= PF_MOVE_DOWN_TEXT
BINDING_NAME_PF_DROP_DOWN	= PF_DROP_DOWN_TEXT

local PF_GEM_TEXTURE_DEFAULTS = {
	-- "nil" (without quotes) indicates not to use a texture for that gem type
	-- Custom textures can be used (eg. "Interface/AddOns/PuzzleFighter/Skins/MyOwnGem")
	-- Custom textures should be 64x64 in size, with type TGA or BLP

	"Interface/Icons/INV_Misc_Gem_Emerald_01",	--  1, Green  Gem
	"Interface/Icons/INV_Misc_Gem_Ruby_01",		--  2, Red    Gem
	"Interface/Icons/INV_Misc_Gem_Sapphire_02",	--  3, Blue   Gem
	"Interface/Icons/INV_Misc_Gem_Topaz_01",	--  4, Yellow Gem
	"Interface/Icons/INV_Potion_91",			--  5, Green  Crash Gem
	"Interface/Icons/INV_Potion_49",			--  6, Red    Crash Gem
	"Interface/Icons/INV_Potion_70",			--  7, Blue   Crash Gem
	"Interface/Icons/INV_Potion_56",			--  8, Yellow Crash Gem
	"Interface/Icons/INV_Misc_Gem_Diamond_01",	--  9, Diamond Crash Gem
	"Interface/Icons/INV_Misc_QuestionMark",	-- 10, Question Mark

	-- These gems do not use textures at the moment
	nil,						-- 11, Green  Counter Gem 1
	nil, 						-- 12, Red    Counter Gem 1
	nil, 						-- 13, Blue   Counter Gem 1
	nil, 						-- 14, Yellow Counter Gem 1
	nil, 						-- 15, Green  Counter Gem 2
	nil, 						-- 16, Red    Counter Gem 2
	nil, 						-- 17, Blue   Counter Gem 2
	nil, 						-- 18, Yellow Counter Gem 2
	nil, 						-- 19, Green  Counter Gem 3
	nil, 						-- 20, Red    Counter Gem 3
	nil, 						-- 21, Blue   Counter Gem 3
	nil, 						-- 22, Yellow Counter Gem 3
	nil, 						-- 23, Green  Counter Gem 4
	nil, 						-- 24, Red    Counter Gem 4
	nil, 						-- 25, Blue   Counter Gem 4
	nil, 						-- 26, Yellow Counter Gem 4
	nil, 						-- 27, Green  Counter Gem 5
	nil, 						-- 28, Red    Counter Gem 5
	nil, 						-- 29, Blue   Counter Gem 5
	nil, 						-- 30, Yellow Counter Gem 5
}

local PF_SOUND_FILE_DEFAULTS = {
	-- Empty string indicates not to play a sound for that event
	LEFT	= "Sound/Character/EmoteClap6.wav",		-- Move Left
	RIGHT	= "Sound/Character/EmoteClap6.wav",		-- Move Right
	DOWN	= "Sound/Character/EmoteClap6.wav",		-- Move Down
	DOWNT	= "Sound/Character/EmoteClap3.wav",		-- Move Down via timer
	ROTATE	= "Sound/Character/EmoteClap6.wav",		-- Rotate
	DROP	= "",									-- Land
	MERGE	= "Sound/interface/PickUp/PutDownGems.wav",	-- Gem Merge
	CRASH	= "Sound/Doodad/G_BarrelOpen-Chest1.wav",	-- Gem Crash
	LOSE	= "Sound/Interface/igQuestFailed.wav",		-- Match Lose
	WIN		= "Sound/Interface/iQuestComplete.wav",		-- Match Win
	ACCEPT	= "Sound/Interface/iQuestActivate.wav",		-- Match Accept
}

local PF_NO_SOUND_FILE_DEFAULTS = {
	LEFT	= "",	-- Move Left
	RIGHT	= "",	-- Move Right
	DOWN	= "",	-- Move Down
	DOWNT	= "",	-- Move Down via timer
	ROTATE	= "",	-- Rotate
	DROP	= "",	-- Land
	MERGE	= "",	-- Gem Merge
	CRASH	= "",	-- Gem Crash
	LOSE	= "",	-- Match Lose
	WIN		= "",	-- Match Win
	ACCEPT	= "",	-- Match Accept
}

-- Note: Must keep "Random" as last option
local PF_CharList = {
	"Ken",
	"Ryu",
	"Chun Li",
	"Sakura",
	"Morrigan",
	"Hsien Ko",
	"Donovan",
	"Felicia",
	"Akuma",
	"Devilot",
	"Dan",
	"Random",
}

-- These gem patterns are defined upside down, the comments are what they look like in game
local PF_GemPatterns = {
	["Ken"] = {
		28, 28, 28, 28, 28, 28,		-- YYYYYY
		27, 27, 27, 27, 27, 27,		-- BBBBBB
		29, 29, 29, 29, 29, 29,		-- GGGGGG
		30, 30, 30, 30, 30, 30},	-- RRRRRR
	["Ryu"] = {
		28, 27, 29, 30, 28, 27,		-- RGBYRG
		28, 27, 29, 30, 28, 27,		-- RGBYRG
		28, 27, 29, 30, 28, 27,		-- RGBYRG
		28, 27, 29, 30, 28, 27},	-- RGBYRG
	["Chun Li"] = {
		28, 28, 27, 27, 29, 29,		-- YYRRGG
		28, 28, 27, 27, 29, 29,		-- YYRRGG
		30, 30, 28, 28, 27, 27,		-- RRGGBB
		30, 30, 28, 28, 27, 27},	-- RRGGBB
	["Sakura"] = {
		27, 28, 28, 28, 28, 30,		-- GBBBBY
		27, 29, 29, 29, 29, 30,		-- GRRRRY
		27, 28, 28, 28, 28, 30,		-- GBBBBY
		27, 29, 29, 29, 29, 30},	-- GRRRRY
	["Morrigan"] = {
		29, 30, 28, 28, 30, 29,		-- YBGGBY
		29, 30, 28, 28, 30, 29,		-- YBGGBY
		30, 29, 27, 27, 29, 30,		-- BYRRYB
		30, 29, 27, 27, 29, 30},	-- BYRRYB
	["Hsien Ko"] = {
		30, 29, 29, 27, 27, 28,		-- GGRRYY
		29, 29, 27, 27, 28, 28,		-- BGGRRY
		29, 27, 27, 28, 28, 30,		-- BBGGRR
		27, 27, 28, 28, 30, 30},	-- YBBGGR
	["Donovan"] = {
		28, 30, 28, 30, 28, 30,		-- GGGBBB
		28, 30, 28, 30, 28, 30,		-- GGGBBB
		27, 27, 27, 29, 29, 29,		-- RYRYRY
		27, 27, 27, 29, 29, 29},	-- RYRYRY
	["Felicia"] = {
		27, 28, 28, 29, 29, 30,		-- GBBRRY
		27, 28, 28, 29, 29, 30,		-- GBBRRY
		27, 29, 29, 28, 28, 30,		-- GRRBBY
		27, 29, 29, 28, 28, 30},	-- GRRBBY
	["Akuma"] = {
		27, 28, 30, 29, 27, 28,		-- RYBGRY
		29, 27, 28, 30, 29, 27,		-- YBGRYB
		30, 29, 27, 28, 30, 29,		-- BGRYBG
		28, 30, 29, 27, 28, 30},	-- GRYBGR
	["Devilot"] = {
		28, 27, 29, 30, 28, 27,		-- YRGBYR
		27, 29, 30, 28, 27, 29,		-- BYRGBY
		29, 30, 28, 27, 29, 30,		-- GBYRGB
		30, 28, 27, 29, 30, 28},	-- RGBYRG
	["Dan"] = {
		28, 28, 28, 28, 28, 28,		-- RRRRRR
		28, 28, 28, 28, 28, 28,		-- RRRRRR
		28, 28, 28, 28, 28, 28,		-- RRRRRR
		28, 28, 28, 28, 28, 28},	-- RRRRRR
	["Random"] = {
		36, 36, 36, 36, 36, 36,		-- ??????
		36, 36, 36, 36, 36, 36,		-- ??????
		36, 36, 36, 36, 36, 36,		-- ??????
		36, 36, 36, 36, 36, 36},	-- ??????
}

local PF_ANIMTIME		= 0.50	-- Time to delay between performing chain animations
local PF_FALLTIME		= 1.00	-- Speed at which gems fall
local PF_MESSAGE_BUFFER_MAX	= 100	-- Max incoming message buffer size

--------- Game Variables -----------

local _
local PF_GamePaused = false
local PF_GameState = 0	-- 0 = Not in game, 1 = Piece is falling, 2 = Piece dropped, process field
local PF_GemString = ""
local PF_IsServer = true
local PF_InitState = 1	-- 1 = Prior to game, 2 = In a game (so game settings can't be changed)
local PF_PauseDelay = 0
PF_OpponentName = PF_NO_OPPONENT_TEXT
PF_ChallengedOpp = PF_NO_OPPONENT_TEXT
local PF_GemTextures = PF_GEM_TEXTURE_DEFAULTS
local PF_SoundFiles = PF_SOUND_FILE_DEFAULTS

local PF_FallingX = 4
local PF_FallingY = 13
local PF_Falling2X = 4
local PF_Falling2Y = 14
local PF_GemDirection = 1	-- 1 = Up, 2 = Right, 3 = Down, 4 = Left

local PF_Hero_P1 = 1
local PF_Hero_P2 = 1
local PF_FallingButton_P1 = 0
local PF_FallingButton2_P1 = 0
local PF_FallingButton_P2 = 0
local PF_FallingButton2_P2 = 0
local PF_Time = 0
local PF_Time2 = 0
local PF_Chain_P1 = 0
local PF_Chain_P2 = 0
local PF_Countdown_P1 = false
local PF_Countdown_P2 = false
local PF_AttackCount_P1 = 0
local PF_AttackedCount_P1 = 0
local PF_AttackedCount_P2 = -1

-- Initialise some arrays
local PF_Field = {}
local PF_ID = {}
local PF_Button = {}

for i = 1, 2 do
	PF_Field[i] = {}
	PF_ID[i] = {}
	PF_Button[i] = {}
	for j = 1, 6 do
		PF_Field[i][j] = {}
		PF_ID[i][j] = {}
	end
end

local PF_MessageBuffer = {}
local PF_MessageBufferStart = 1
local PF_MessageBufferEnd = 1

--------- Localized Globals -----------

local strsub, strfind, random = strsub, strfind, random

--------- Saved Variables -----------

PF_Options = {}

--------- Local Game Functions -----------

function PF_SendMsg(msg)
	-- Send message to opponent only if there is an opponent
	if PF_OpponentName ~= PF_NO_OPPONENT_TEXT then
		SendAddonMessage("PuzzleFighter", msg, "WHISPER", PF_OpponentName)
	end
end

function PF_PauseGame(byOpponent)
	if PF_InitState == 2 then
		PF_CellContainer_P1:SetScript("OnUpdate", nil)
		PF_PauseButton:LockHighlight()
		PF_PauseButton:SetText(PF_UNPAUSE_TEXT)
		PF_PauseDelay = GetTime() - PF_Time
		PF_GamePaused = true
		if IsShiftKeyDown() and IsControlKeyDown() then
		else
			if byOpponent ~= true then PF_SendMsg("PAUS pause") end
		end
	end
end

function PF_UnpauseGame(byOpponent)
	if PF_InitState == 2 then
		PF_Time = GetTime() - PF_PauseDelay
		PF_CellContainer_P1:SetScript("OnUpdate", PF_OnUpdate)
		PF_PauseButton:UnlockHighlight()
		PF_PauseButton:SetText(PF_PAUSE_TEXT)
		PF_PauseDelay = 0
		PF_GamePaused = false
		if IsShiftKeyDown() and IsControlKeyDown() then
		else
			if byOpponent ~= true then PF_SendMsg("PAUS unpause") end
		end
	end
end

function PF_GenerateGemString()
	-- Generates a string of 50 1-digit numbers (gem generation)
	-- Gem 49 is always a diamond (first gem on turn 25), so every 25
	-- turns a diamond will spawn
	local a, b
	b = ""

	-- First 2 gems will always be normal gems to prevent Turn 1 empty field bonus.
	for i = 1, 2 do
		a = random(1, 4)
		b = b..a
	end

	for i = 3, 48 do
		-- 1/4 chance to be a crash gem
		a = random(1, 8)
		if a >= 5 then
			a = random(1, 8)
		end
		b = b..a
	end

	-- Gem 49 is always a diamond
	b = b.."9"

	-- Gem 50 -- Pair the diamond with a normal gem
	a = random(1, 4)
	b = b..a

	return b
end

function PF_GetNextGem()
	-- Extracts the first digit out of the gem string and returns it
	local a = tonumber(strsub(PF_GemString, 1, 1))
	PF_GemString = strsub(PF_GemString, 2)
	return a
end

function PF_GetNextGemPreview()
	return tonumber(strsub(PF_GemString, 1, 1)), tonumber(strsub(PF_GemString, 2, 2))
end

function PF_UpdateNextGem(player, a, b)
	-- Updates the display of the next pair of gems
	if player == 1 then
		PF_Next1_P1:SetNormalTexture(PF_GemTextures[a])
		PF_Next2_P1:SetNormalTexture(PF_GemTextures[b])
	else
		PF_Next1_P2:SetNormalTexture(PF_GemTextures[a])
		PF_Next2_P2:SetNormalTexture(PF_GemTextures[b])
	end
end

function PF_GetUnusedButton(player)
	-- Find first unused button and return it
	for i = 1, 86 do
		if PF_Button[player][i].gem == 0 then
			return PF_Button[player][i]
		end
	end
end

function PF_DrawCell(button, X, Y)
	button.CurrentX = X*32-22
	button.CurrentY = Y*32-22
	button.TargetX = X*32-22
	button.TargetY = Y*32-22
	button:SetPoint("BOTTOMLEFT", button.container, "BOTTOMLEFT", X*32-22, Y*32-22)
	button:SetNormalTexture(PF_GemTextures[button.gem])
	button:Show()
end

function PF_HideCell(button)
	--local a = PF_Options["Transparency"]
	--if a == nil then a = 100 end

	button.gem = 0
	button:Hide()
	button:ClearAllPoints()
	button:SetAlpha(1)
	button:SetScript("OnUpdate", nil)
	button.border:Hide()
	button.label:Hide()
end

function PF_FadeCell(this, elapsed)
	-- Throtle to 25 fps
	this.timer = (this.timer or 0) + elapsed
	if this.timer < 0.04 then return end
	this.timer = 0

	local a = this:GetAlpha() - 0.08
	if a <= 0 then
		PF_HideCell(this)
	else
		this:SetAlpha(a)
	end
end

function PF_ButtonMove(this, elapsed)
	-- Throtle to 100 fps
	this.timer = (this.timer or 0) + elapsed
	if this.timer < 0.01 then return end
	this.timer = 0

	local speed = 8

	if this.CurrentX > this.TargetX then
		this.CurrentX = this.CurrentX - speed
	end
	if this.CurrentY > this.TargetY then
		this.CurrentY = this.CurrentY - speed
	end
	if this.CurrentX < this.TargetX then
		this.CurrentX = this.CurrentX + speed
	end
	if this.CurrentY < this.TargetY then
		this.CurrentY = this.CurrentY + speed
	end

	this:SetPoint("BOTTOMLEFT", this.container, "BOTTOMLEFT", this.CurrentX, this.CurrentY)

	if this.CurrentX == this.TargetX and this.CurrentY == this.TargetY then
		this:SetScript("OnUpdate", nil)
	end
end

function PF_SpawnGem()
	local player = 1

	-- Gems start in column 4, row 14
	PF_FallingX = 4
	PF_FallingY = 14

	if PF_Field[player][PF_FallingX][PF_FallingY] ~= 0 then
		-- Game over
		PlaySoundFile(PF_SoundFiles["LOSE"])

		PF_CellContainer_P1:SetScript("OnUpdate", nil)
		PF_StatusLabel:SetText(PF_STATUS_TEXT..PF_GAME_OVER_LOST_TEXT)
		PF_PauseButton:SetText(START)
		PF_GameState = 0
		PF_InitState = 1
		PF_SendMsg("GOVR")
		PF_ClearMessageBuffer()
		PF_Frame:SetScript("OnUpdate", nil)

		PF_ChainText_P1:SetTextColor(1, 0, 0, 1)
		PF_ChainText_P1:SetText(PF_YOU_LOSE_TEXT)
		PF_ChainFrame_P1:SetPoint("CENTER", PF_CellContainer_P1, "CENTER", 0, -24)
		PF_ChainFrame_P1:SetAlpha(1.0)
		PF_ChainFrame_P1:SetScript("OnUpdate", PF_ChainFrameOnUpdate)

		if PF_OpponentName ~= PF_NO_OPPONENT_TEXT then
			PF_ChainText_P2:SetTextColor(0, 1, 0, 1)
			PF_ChainText_P2:SetText(PF_WINNER_TEXT)
			PF_ChainFrame_P2:SetPoint("CENTER", PF_CellContainer_P2, "CENTER", 0, -24)
			PF_ChainFrame_P2:SetAlpha(1.0)
			PF_ChainFrame_P2:SetScript("OnUpdate", PF_ChainFrameOnUpdate)
		end

		PF_OpponentName = PF_NO_OPPONENT_TEXT
		PF_ChallengedOpp = PF_NO_OPPONENT_TEXT

		return
	end

	-- Check PF_GemString to see if its almost empty
	if #PF_GemString <= 10 then
		if PF_IsServer == true then
			-- It is, so populate it with the next 30 random pairs of gems
			local a = PF_GenerateGemString()
			PF_GemString = PF_GemString..a
			PF_SendMsg("GSTR "..a)
		else
			PF_SendMsg("GREQ")
		end
	end

	-- Get an unused button and update that button on screen for Gem 1
	PF_FallingButton_P1 = PF_GetUnusedButton(1)
	PF_FallingButton_P1.gem = PF_GetNextGem()
	PF_FallingButton_P1.SizeX = 1
	PF_FallingButton_P1.SizeY = 1
	PF_FallingButton_P1:SetWidth(32)
	PF_FallingButton_P1:SetHeight(32)
	PF_DrawCell(PF_FallingButton_P1, PF_FallingX, PF_FallingY)


	-- Attached gem start in column 4, row 15
	PF_Falling2X = 4
	PF_Falling2Y = 15
	PF_GemDirection = 1  -- Above the other gem

	-- Get another unused button and update that button on screen for Gem 2
	PF_FallingButton2_P1 = PF_GetUnusedButton(1)
	PF_FallingButton2_P1.gem = PF_GetNextGem()
	PF_FallingButton2_P1.SizeX = 1
	PF_FallingButton2_P1.SizeY = 1
	PF_FallingButton2_P1:SetWidth(32)
	PF_FallingButton2_P1:SetHeight(32)
	PF_DrawCell(PF_FallingButton2_P1, PF_Falling2X, PF_Falling2Y)

	local a, b
	a, b = PF_GetNextGemPreview()
	PF_SendMsg("SPWN "..PF_FallingButton_P1.gem..":"..PF_FallingButton2_P1.gem..":"..a..":"..b)

	PF_UpdateNextGem(1, a, b)
	PF_GameState = 1
	PF_Chain_P1 = 0
end

function PF_StartGame()
	PF_InitState = 2

	if PF_OpponentName == PF_NO_OPPONENT_TEXT then
		PF_StatusLabel:SetText(PF_STATUS_TEXT..PF_PRACTICE_GAME_TEXT)
	else
		PF_StatusLabel:SetText(PF_STATUS_TEXT..format(PF_MATCH_IN_PROGRESS_TEXT, PF_ChallengedOpp))
	end

	PF_PauseButton:UnlockHighlight()
	PF_PauseButton:SetText(PF_PAUSE_TEXT)
	PF_PauseButton:Enable()
	PF_CharDropDown_P1:Hide()
	PF_CharDropDown_P2:Hide()
	PF_ReadyButton:Hide()

	-- If current hero selection is the last one (random selection),
	-- random a hero and update opponent
	if PF_Hero_P1 == #PF_CharList then
		PF_Hero_P1 = random(1, #PF_CharList - 1)
		PF_ShowPattern(1, PF_Hero_P1)
		PF_CharPatternFrameLabel_P1:SetText("<-- "..PF_CharList[PF_Hero_P1])
		PF_SendMsg("RCHR "..PF_Hero_P1)
	end

	-- Spawn initial gem
	PF_SpawnGem()
	PF_Time = GetTime()
	PF_CellContainer_P1:SetScript("OnUpdate", PF_OnUpdate)
end

function PF_NewGame()
	PF_Reset(true)

	if PF_OpponentName == PF_NO_OPPONENT_TEXT or PF_OpponentName == UnitName("player") then
		-- Start single player game (or game against yourself)
		PF_IsServer = true
		PlaySoundFile(PF_SoundFiles["ACCEPT"])
		PF_CharDropDown_P1:Show()
		PF_CharDropDown_P2:Show()
		PF_CharPatternFrameLabel_P1:Show()
		PF_CharPatternFrameLabel_P2:Show()
		PF_ReadyButton:Enable()
		PF_ReadyButton:Show()
	end

	if PF_OpponentName == PF_NO_OPPONENT_TEXT then
		-- Single player game (no opponent)
		PF_CharPatternFrame_P1:Hide()
		PF_CharPatternFrame_P2:Hide()
		PF_CharPatternFrameLabel_P1:Hide()
		PF_CharPatternFrameLabel_P2:Hide()
		PF_StartGame()
	else
		PF_Frame:SetScript("OnUpdate", PF_ProcessMessage)
	end
end

-- Function to check that cell (i, j) is size 1x1 and of a certain colour
function PF_Check1x1(player, i, j, color)
	-- Check colour
	if PF_Field[player][i][j] ~= color then
		return false
	end

	-- Now check for 1x1
	local a = PF_ID[player][i][j]
	if a <= 0 then
		return false  -- invalid button, cell is part of another power gem
	end

	a = PF_Button[player][a]
	if a.SizeX == 1 and a.SizeY == 1 then
		return true   -- cell is 1x1 and matches colour
	else
		return false
	end
end

function PF_CheckCoalesce(player)
	-- This function attempts to find and coalesce smaller gems into larger ones
	-- It gives priority to existing gems coalescing to even larger gems first before
	-- checking for normal gems that can possibly coalesce into a power gem.

	-- At any one time a coalesce is successful, the function exits and returns true
	-- so that the calling function can call this function again for further coalescing,
	-- because sometimes a smaller coalesce makes another larger coalesce possible.
	-- Once no more coalescing is possible, the function exits and returns false.

	-- Find 2x2 or larger gems and try to enlarge them into even larger gems first
	-- Perform the j loop first instead of the i loop so that we scan from bottom to top
	local a, b, c, d

	for j = 1, 13 do
		for i = 1, 5 do
			-- Check that the square contains a gem (gem 1 to 4)
			-- and that it has a button ID
			a = PF_Field[player][i][j]
			b = PF_ID[player][i][j]
			if a > 0 and a < 5 and b > 0 then
				-- Now get the button and check that it is a power gem
				b = PF_Button[player][b]
				if b.SizeX > 1 and b.SizeY > 1 then
					-- This is the lower left corner of a power gem

					-- Check for extension to the left
					if i > 1 then
						-- Check that all of them are 1x1s and
						-- are of the same colour
						local flag = true
						for k = j, (j + b.SizeY - 1) do
							if PF_Check1x1(player, i - 1, k, a) == false then
								flag = false
							end
						end
						if flag then
							-- First hide the affected gems and update the ID fields
							for k = j, (j + b.SizeY - 1) do
								c = PF_ID[player][i - 1][k]
								c = PF_Button[player][c]
								PF_HideCell(c)
								PF_ID[player][i - 1][k] = 0 - PF_ID[player][i][j]
							end
							-- Move the button left 1 column then extend it by 1 column
							PF_ID[player][i - 1][j] = PF_ID[player][i][j]
							PF_ID[player][i][j] = 0 - PF_ID[player][i][j]
							b.CurrentX = b.CurrentX - 32
							b.TargetX = b.TargetX - 32
							b.SizeX = b.SizeX + 1
							b:SetWidth(b.SizeX * 32)
							b:SetPoint("BOTTOMLEFT", b.container, "BOTTOMLEFT", b.CurrentX, b.CurrentY)
							return true
						end
					end

					-- Check for extension to the right
					if i + b.SizeX < 7 then
						-- Check that all of them are 1x1s and
						-- are of the same colour
						local flag = true
						for k = j, (j + b.SizeY - 1) do
							if PF_Check1x1(player, i + b.SizeX, k, a) == false then
								flag = false
							end
						end
						if flag then
							-- First hide the affected gems and update the ID fields
							for k = j, (j + b.SizeY - 1) do
								c = PF_ID[player][i + b.SizeX][k]
								c = PF_Button[player][c]
								PF_HideCell(c)
								PF_ID[player][i + b.SizeX][k] = 0 - PF_ID[player][i][j]
							end
							-- Extend the actual button by 1 column
							b.SizeX = b.SizeX + 1
							b:SetWidth(b.SizeX * 32)
							return true
						end
						-- Alternatively, check that the power gem to the
						-- right (if one) is of the same height and colour
						c = PF_Field[player][i + b.SizeX][j]
						d = PF_ID[player][i + b.SizeX][j]
						if c == a and d > 0 then
							-- Now get the button and check that it is a power gem
							-- of the same height
							d = PF_Button[player][d]
							if d.SizeY == b.SizeY then
								-- It is, so merge the 2 power gems
								-- First hide the affected power gem and update the ID fields
								PF_HideCell(d)
								for k = (i + b.SizeX), (i + b.SizeX + d.SizeX - 1) do
								for l = j, (j + d.SizeY - 1) do
									PF_ID[player][k][l] = 0 - PF_ID[player][i][j]
								end
								end
								-- Now extend the actual button
								b.SizeX = b.SizeX + d.SizeX
								b:SetWidth(b.SizeX * 32)
								return true
							end
						end
					end

					-- Check for extension above
					if j + b.SizeY < 15 then
						-- Check that all of them are 1x1s and
						-- are of the same colour
						local flag = true
						for k = i, (i + b.SizeX - 1) do
							if PF_Check1x1(player, k, j + b.SizeY, a) == false then
								flag = false
							end
						end
						if flag then
							-- First hide the affected gems and update the ID fields
							for k = i, (i + b.SizeX - 1) do
								c = PF_ID[player][k][j + b.SizeY]
								c = PF_Button[player][c]
								PF_HideCell(c)
								PF_ID[player][k][j + b.SizeY] = 0 - PF_ID[player][i][j]
							end
							-- Extend the actual button by 1 column
							b.SizeY = b.SizeY + 1
							b:SetHeight(b.SizeY * 32)
							return true
						end
						-- Alternatively, check that the power gem to the
						-- top (if one) is of the same width and colour
						c = PF_Field[player][i][j + b.SizeY]
						d = PF_ID[player][i][j + b.SizeY]
						if c == a and d > 0 then
							-- Now get the button and check that it is a power gem
							-- of the same height
							d = PF_Button[player][d]
							if d.SizeX == b.SizeX then
								-- It is, so merge the 2 power gems
								-- First hide the affected power gem and update the ID fields
								PF_HideCell(d)
								for k = (i), (i + d.SizeX - 1) do
								for l = (j + b.SizeY), (j + b.SizeY + d.SizeY - 1) do
									PF_ID[player][k][l] = 0 - PF_ID[player][i][j]
								end
								end
								-- Now extend the actual button
								b.SizeY = b.SizeY + d.SizeY
								b:SetHeight(b.SizeY * 32)
								return true
							end
						end

					end

					-- Check for extension below
					if j > 1 then
						-- Check that all of them are 1x1s and
						-- are of the same colour
						local flag = true
						for k = i, (i + b.SizeX - 1) do
							if PF_Check1x1(player, k, j - 1, a) == false then
								flag = false
							end
						end
						if flag then
							-- First hide the affected gems and update the ID fields
							for k = i, (i + b.SizeX - 1) do
								c = PF_ID[player][k][j - 1]
								c = PF_Button[player][c]
								PF_HideCell(c)
								PF_ID[player][k][j - 1] = 0 - PF_ID[player][i][j]
							end
							-- Move the button down 1 row then extend it up by 1 row
							PF_ID[player][i][j - 1] = PF_ID[player][i][j]
							PF_ID[player][i][j] = 0 - PF_ID[player][i][j]
							b.CurrentY = b.CurrentY - 32
							b.TargetY = b.TargetY - 32
							b.SizeY = b.SizeY + 1
							b:SetHeight(b.SizeY * 32)
							b:SetPoint("BOTTOMLEFT", b.container, "BOTTOMLEFT", b.CurrentX, b.CurrentY)
							return true
						end
					end

					-- Checks done (in all 4 directions) for this power gem
				end
			end
		end
	end

	-- Code to combine four 1x1 gems into a 2x2 power gem
	for j = 1, 13 do
		for i = 1, 5 do
			-- Check that 4 squares are the same gem (gem 1 to 4)
			a = PF_Field[player][i][j]
			b = PF_Field[player][i+1][j]
			c = PF_Field[player][i][j+1]
			d = PF_Field[player][i+1][j+1]
			if a > 0 and a < 5 and a == b and a == c and a == d then
				-- Check that all 4 have different button IDs
				a = PF_ID[player][i][j]
				b = PF_ID[player][i+1][j]
				c = PF_ID[player][i][j+1]
				d = PF_ID[player][i+1][j+1]
				if a > 0 and b > 0 and c > 0 and d > 0 then
				if a ~= b and a ~= c and a ~= d and b ~= c and b ~= d and c ~= d then
					a = PF_Button[player][a]
					b = PF_Button[player][b]
					c = PF_Button[player][c]
					d = PF_Button[player][d]
					if a.SizeX == 1 and b.SizeX == 1 and c.SizeX == 1 and d.SizeX == 1 then
					if a.SizeY == 1 and b.SizeY == 1 and c.SizeY == 1 and d.SizeY == 1 then
						-- We have found 4 gems that are all 1x1 in size of the same colour
						-- Expand button A and hide B, C, D
						a:SetWidth(64)
						a:SetHeight(64)
						a.SizeX = 2
						a.SizeY = 2
						PF_HideCell(b)
						PF_HideCell(c)
						PF_HideCell(d)
						-- Update the internal field button IDs
						-- Use negative IDs to indicate these cells are not the primary cell.
						PF_ID[player][i+1][j] = 0 - PF_ID[player][i][j]
						PF_ID[player][i][j+1] = 0 - PF_ID[player][i][j]
						PF_ID[player][i+1][j+1] = 0 - PF_ID[player][i][j]
						return true
					end
					end
				end
				end
			end
		end
	end

	-- No coalescing is performed so far if the function reaches this point
	return false
end

-- Recursive function that breadth-first searches for adjacent gems to break
-- The 4th argument a is the crash gem colour index
function PF_MarkGemBurst(player, i, j, a)
	local b, c, bonus
	bonus = 0
	b = 100

	-- Calculate bonus of this gem (if there is a button for it)
	c = PF_ID[player][i][j]
	if c > 0 then
		c = PF_Button[player][c]
		if c.gem > 0 and c.gem < 5 then
			-- Find the smallest dimension of this gem
			-- Bonus is smallest-dimension * size of gem
			if c.SizeX < b then b = c.SizeX end
			if c.SizeY < b then b = c.SizeY end
			bonus = bonus + b * c.SizeX * c.SizeY
		elseif c.gem > 4 and c.gem < 9 then
			-- This is a crash gem, worth 1 point
			bonus = bonus + 1
		end
	end

	-- Change the internal color field to negative to imply that cell is breaking
	PF_Field[player][i][j] = 0 - PF_Field[player][i][j]

	-- Recurse up, down, left and right to look for adjacent cells to break
	if j < 14 and PF_Field[player][i][j+1] == a-4 then bonus = bonus + PF_MarkGemBurst(player, i, j+1, a) end
	if j > 1 and PF_Field[player][i][j-1] == a-4 then bonus = bonus + PF_MarkGemBurst(player, i, j-1, a) end
	if i > 1 and PF_Field[player][i-1][j] == a-4 then bonus = bonus + PF_MarkGemBurst(player, i-1, j, a) end
	if i < 6 and PF_Field[player][i+1][j] == a-4 then bonus = bonus + PF_MarkGemBurst(player, i+1, j, a) end

	-- Dont' forget crash gems of the same colour too
	if j < 14 and PF_Field[player][i][j+1] == a then bonus = bonus + PF_MarkGemBurst(player, i, j+1, a) end
	if j > 1 and PF_Field[player][i][j-1] == a then bonus = bonus + PF_MarkGemBurst(player, i, j-1, a) end
	if i > 1 and PF_Field[player][i-1][j] == a then bonus = bonus + PF_MarkGemBurst(player, i-1, j, a) end
	if i < 6 and PF_Field[player][i+1][j] == a then bonus = bonus + PF_MarkGemBurst(player, i+1, j, a) end

	return bonus
end

-- This function checks for gems to burst. It marks gems to burst if any.
-- Returns attack bonus value.
function PF_CheckDiamondGemBurst(player)
	local a, bonus
	bonus = 0

	-- Check for a diamond crash gem first
	for j = 1, 14 do
		for i = 1, 6 do
			a = PF_Field[player][i][j]
			if a == 9 then
				-- There is a diamond crash gem here, so check the
				-- colour of the gem below it
				if j > 1 then
					a = PF_Field[player][i][j-1]
					if a > 4 and a < 9 then
						-- Crash gem colour, deduct 4 to get normal gem colour
						a = a - 4
					end
					if a > 10 and a < 31 then
						-- Counter gem colour, add 1, mod 4 then add 1 again to get normal gem colour
						a = (a + 1) % 4 + 1
					end

					-- Now loop through the whole field and mark all crash/counter/gems of
					-- that colour for bursting (change the colour field to negative)
					for l = 1, 14 do
						for k = 1, 6 do
							if PF_Field[player][k][l] == a then PF_Field[player][k][l] = 0 - PF_Field[player][k][l] bonus = bonus + 1 end
							if PF_Field[player][k][l] == a+4 then PF_Field[player][k][l] = 0 - PF_Field[player][k][l] bonus = bonus + 1 end
							if PF_Field[player][k][l] == a+10 then PF_Field[player][k][l] = 0 - PF_Field[player][k][l] bonus = bonus + 0.5 end
							if PF_Field[player][k][l] == a+14 then PF_Field[player][k][l] = 0 - PF_Field[player][k][l] bonus = bonus + 0.5 end
							if PF_Field[player][k][l] == a+18 then PF_Field[player][k][l] = 0 - PF_Field[player][k][l] bonus = bonus + 0.5 end
							if PF_Field[player][k][l] == a+22 then PF_Field[player][k][l] = 0 - PF_Field[player][k][l] bonus = bonus + 0.5 end
							if PF_Field[player][k][l] == a+26 then PF_Field[player][k][l] = 0 - PF_Field[player][k][l] bonus = bonus + 0.5 end
						end
					end
				else
					-- Diamond Crash Gem on row 1, no colour is affected
					-- Award a tech bonus
					bonus = bonus + 12

					if player == 1 then
						PF_TechFrameText_P1:SetTextColor(1, 1, 1, 1)
						PF_TechFrameText_P1:SetText(PF_TECH_BONUS_TEXT)
						PF_TechFrame_P1:SetPoint("BOTTOM", PF_CellContainer_P1, "BOTTOMLEFT", i * 32 - 6, 16)
						PF_TechFrame_P1:SetAlpha(1.0)
						PF_TechFrame_P1:SetScript("OnUpdate", PF_TechFrameOnUpdate)
					else
						PF_TechFrameText_P2:SetTextColor(1, 1, 1, 1)
						PF_TechFrameText_P2:SetText(PF_TECH_BONUS_TEXT)
						PF_TechFrame_P2:SetPoint("BOTTOM", PF_CellContainer_P2, "BOTTOMLEFT", i * 32 - 6, 16)
						PF_TechFrame_P2:SetAlpha(1.0)
						PF_TechFrame_P2:SetScript("OnUpdate", PF_TechFrameOnUpdate)
					end
				end

				-- Finally, mark itself for bursting and return since there won't
				-- be a second diamond crash gem in the field
				PF_Field[player][i][j] = 0 - PF_Field[player][i][j]
				return bonus
			end
		end
	end

	-- No diamond crash gem was found
	return 0
end

-- This function checks for gems to burst. It marks gems to burst if any.
-- Returns attack bonus value.
function PF_CheckGemBurst(player)
	local a, bonus
	bonus = PF_CheckDiamondGemBurst(player)

	for j = 1, 14 do
		for i = 1, 6 do
			a = PF_Field[player][i][j]
			if a > 4 and a < 9 then
				-- This is a Crash Gem, so check to see if there are
				-- gems of the same colour to break

				local flag2 = false
				-- Check up, down, left and right for at least one gem to break...
				if j < 14 and PF_Field[player][i][j+1] == a-4 then flag2 = true end
				if j > 1 and PF_Field[player][i][j-1] == a-4 then flag2 = true end
				if i > 1 and PF_Field[player][i-1][j] == a-4 then flag2 = true end
				if i < 6 and PF_Field[player][i+1][j] == a-4 then flag2 = true end
				-- Or check for another crash gem of the same colour...
				if j < 14 and PF_Field[player][i][j+1] == a then flag2 = true end
				if j > 1 and PF_Field[player][i][j-1] == a then flag2 = true end
				if i > 1 and PF_Field[player][i-1][j] == a then flag2 = true end
				if i < 6 and PF_Field[player][i+1][j] == a then flag2 = true end

				if flag2 then
					bonus = bonus + PF_MarkGemBurst(player, i, j, a)
				end
			end
		end
	end

	-- Now check for counter gems that are next to a normal gem that is breaking
	for j = 1, 14 do
		for i = 1, 6 do
			a = PF_Field[player][i][j]
			if a > 10 and a < 31 then
				local flag2 = false
				-- This is a countergem, so check if there is
				-- a crash gem next to it that is breaking
				if j < 14 and PF_Field[player][i][j+1] < 0 and PF_Field[player][i][j+1] > -9 then flag2 = true end
				if j > 1 and PF_Field[player][i][j-1] < 0 and PF_Field[player][i][j-1] > -9 then flag2 = true end
				if i > 1 and PF_Field[player][i-1][j] < 0 and PF_Field[player][i-1][j] > -9 then flag2 = true end
				if i < 6 and PF_Field[player][i+1][j] < 0 and PF_Field[player][i+1][j] > -9 then flag2 = true end

				if flag2 then
					-- There is one, so break this counter gem
					PF_Field[player][i][j] = 0 - PF_Field[player][i][j]
					bonus = bonus + 0.5
				end
			end
		end
	end

	return bonus
end

function PF_PerformGemBurst(player)
	local a
	for j = 1, 14 do
		for i = 1, 6 do
			if PF_Field[player][i][j] < 0 then
				-- This is a gem that is breaking, so check if there's
				-- a button for it
				a = PF_ID[player][i][j]
				if a > 0 then
					-- There is a button, so fade it out
					a = PF_Button[player][a]
					a:SetScript("OnUpdate", PF_FadeCell)
				end

				-- And clear the internal fields
				PF_ID[player][i][j] = 0
				PF_Field[player][i][j] = 0
			end
		end
	end
end

function PF_CollapseField(player)
	-- Search for gems to drop down starting from row 2
	local a
	for j = 2, 14 do
		for i = 1, 6 do
			a = PF_ID[player][i][j]
			if a > 0 then
				-- There is a gem here, now check if it can fall by 1 row
				local flag
				a = PF_Button[player][a]

				for l = j-1, 1, -1 do
					flag = true
					for k = i, (i + a.SizeX - 1) do
						if PF_Field[player][k][l] ~= 0 then
							flag = false
						end
					end

					if flag then
						-- All cells are empty in the row below this gem
						-- So move the button down by 1 row
						a.TargetY = l * 32 - 22
						a:SetScript("OnUpdate", PF_ButtonMove)

						-- Update the internal fields
						for m = (l + 1), (l + a.SizeY) do
							for k = i, (i + a.SizeX - 1) do
								PF_Field[player][k][m-1] = PF_Field[player][k][m]
								PF_ID[player][k][m-1] = PF_ID[player][k][m]
							end
						end
						for k = i, (i + a.SizeX - 1) do
							PF_Field[player][k][l + a.SizeY] = 0
							PF_ID[player][k][l + a.SizeY] = 0
						end
					else
						-- Can't fall, so stop and check next gem
						break
					end
				end
			end
		end
	end
end

function PF_ChainFrameOnUpdate(this, elapsed)
	-- Throtle to 25 fps
	this.timer = (this.timer or 0) + elapsed
	if this.timer < 0.04 then return end
	this.timer = 0

	local speed = 2
	if PF_GameState == 0 and PF_InitState == 1 then
		-- This condition is true when a "Game Over" is triggered
		-- or in the initialisation stage before a game starts
		speed = 1
	end

	-- Move the message up over time
	local point, relativeFrame, relativeFPoint, offsetX, offsetY = this:GetPoint("CENTER")
	if offsetY < 24 then
		this:SetPoint(point, relativeFrame, relativeFPoint, offsetX, offsetY + speed)
	else
		this:SetScript("OnUpdate", nil)
	end

	-- Fade the message out over time (after it moves above the center)
	if offsetY >= 0 then
		this:SetAlpha(1 - (offsetY + speed)/24)
	end
end

function PF_TechFrameOnUpdate(this, elapsed)
	-- Throtle to 25 fps
	this.timer = (this.timer or 0) + elapsed
	if this.timer < 0.04 then return end
	this.timer = 0

	local speed = 2

	-- Move the message up over time
	local point, relativeFrame, relativeFPoint, offsetX, offsetY = this:GetPoint("BOTTOM")
	if offsetY < 64 then
		this:SetPoint(point, relativeFrame, relativeFPoint, offsetX, offsetY + speed)
	else
		this:SetScript("OnUpdate", nil)
	end

	-- Fade the message out over time (after it moves above the center)
	if offsetY >= 40 then
		this:SetAlpha(1 - (offsetY + speed - 40)/24)
	end
end

-- Return the counter gem colour (27-30) of the opponent's hero
-- and column to drop it in, given the attack count.
function PF_GetCounterColor(player, count)
	local hero, color, column
	if player == 1 then
		hero = PF_CharList[PF_Hero_P2]
	else
		hero = PF_CharList[PF_Hero_P1]
	end

	column = count % 6
	if column >= 4 then
		column = column + 1
		color = PF_GemPatterns[hero][count % 24 + 1]
	elseif column == 0 then
		column = 4
		color = PF_GemPatterns[hero][(count - 3) % 24 + 1]
	else
		color = PF_GemPatterns[hero][(count - 1) % 24 + 1]
	end

	return color, column
end

-- Add a counter gem to "player" in "column" of "colour" (valid colors 27-30)
function PF_AddCounterGem(player, column, color, height)
	local i, row, button, border, label
	if not height then height = 0 end

	row = 1
	for i = 15, 2, -1 do
		if PF_Field[player][column][i-1] ~= 0 then
			row = i
			break
		end
	end

	if row < 15 then
		PF_Field[player][column][row] = color
		button = PF_GetUnusedButton(player)
		PF_ID[player][column][row] = button:GetID()
		button.gem = color
		button.SizeX = 1
		button.SizeY = 1
		button:SetWidth(32)
		button:SetHeight(32)

		border = button.border
		label = button.label
		label:SetText("5")
		if color == 27 then			-- Green
			label:SetTextColor(0, 1, 0, 1)
			border:SetVertexColor(0, 1, 0, 1)
		elseif color == 28 then		-- Red
			label:SetTextColor(1, 0, 0, 1)
			border:SetVertexColor(1, 0, 0, 1)
		elseif color == 29 then		-- Blue
			label:SetTextColor(0.3, 0.3, 1, 1)
			border:SetVertexColor(0.3, 0.3, 1, 1)
		elseif color == 30 then		-- Yellow
			label:SetTextColor(1, 1, 0, 1)
			border:SetVertexColor(1, 1, 0, 1)
		end
		if PF_GemTextures[button.gem] == nil then
			-- Show only if there is no texture
			label:Show()
			border:Show()
		else
			button:SetNormalTexture(PF_GemTextures[button.gem])
		end

		PF_DrawCell(button, column, 14 + height)
		button.TargetX = column*32-22
		button.TargetY = row*32-22
		button:SetScript("OnUpdate", PF_ButtonMove)
	end
end

-- Add "count" counter gems to "player" field.
function PF_Attack(player, count)
	local color, column

	for i = 1, count do
		color, column = PF_GetCounterColor(player, i)
		PF_AddCounterGem(player, column, color, math.floor((i-1)/6))
	end
end

-- Counts down all counter gems and convert "1"s to normal gems
function PF_CountdownCounterGems(player)
	local a, b, label

	for j = 1, 14 do
		for i = 1, 6 do
			a = PF_Field[player][i][j]
			if a > 10 and a < 15 then
				b = PF_ID[player][i][j]
				b = PF_Button[player][b]

				-- Do this portion only if there is no texture
				b.border:Hide()
				label = b.label
				label:SetText("")
				label:Hide()

				b.gem = b.gem - 10
				PF_Field[player][i][j] = PF_Field[player][i][j] - 10
				b:SetNormalTexture(PF_GemTextures[b.gem])
			elseif a > 14 and a < 31 then
				b = PF_ID[player][i][j]
				b = PF_Button[player][b]

				-- Do this portion only if there is no texture
				label = b.label
				label:SetText(label:GetText() - 1)

				b.gem = b.gem - 4
				PF_Field[player][i][j] = PF_Field[player][i][j] - 4
				b:SetNormalTexture(PF_GemTextures[b.gem])
			end
		end
	end
end

-- Check if a player's field is completely empty (so only need to check row 1)
function PF_CheckEmptyField(player)
	local flag = true
	for i = 1, 6 do
		if PF_Field[player][i][1] ~= 0 then flag = false end
	end
	return flag
end

function PF_OnUpdate()
	local a, b, bonus
	if PF_GameState == 1 then
		a = PF_FALLTIME
	else
		if PF_FallingButton_P1:GetScript("OnUpdate") then return end
		if PF_FallingButton2_P1:GetScript("OnUpdate") then return end
		a = PF_ANIMTIME
	end

	if GetTime() - PF_Time >= a then
		PF_Time = PF_Time + a

		if PF_GameState == 1 then
			PF_MoveDown(true)
		elseif PF_GameState == 2 then
			PF_RemoveRow15(1)
			-- Countdown the counter gems once
			if PF_Countdown_P1 == true then
				PF_CountdownCounterGems(1)
				PF_Countdown_P1 = false
			end

			-- A) Perform gem coalescing repeatedly until none can be performed (empty loop)
			b = false
			while PF_CheckCoalesce(1) == true do b = true end
			if b then
				PlaySoundFile(PF_SoundFiles["MERGE"])
			end

			-- B) Check for gem burst
			bonus = PF_CheckGemBurst(1)
			if bonus > 0 then
				-- Every chain above the first one gives a cumulative 10% bonus for that chain
				PF_AttackCount_P1 = PF_AttackCount_P1 + bonus * (1 + PF_Chain_P1 / 10)

				-- C) Perform gem burst and collapse the field
				PF_PerformGemBurst(1)
				PF_GameState = 3
				PlaySoundFile(PF_SoundFiles["CRASH"])

				-- Display chains above 1
				PF_Chain_P1 = PF_Chain_P1 + 1
				if PF_Chain_P1 > 1 then
					PF_ChainText_P1:SetTextColor(1, 0, 0, 1)
					PF_ChainText_P1:SetText(PF_Chain_P1..PF_CHAIN_TEXT)
					PF_ChainFrame_P1:SetPoint("CENTER", PF_CellContainer_P1, "CENTER", 0, -24)
					PF_ChainFrame_P1:SetAlpha(1.0)
					PF_ChainFrame_P1:SetScript("OnUpdate", PF_ChainFrameOnUpdate)
				end

				-- Go to (D) after animation of gem burst is done (via this OnUpdate function)
			else
				-- Check for empty field bonus
				if PF_CheckEmptyField(1) == true then
					PF_AttackCount_P1 = PF_AttackCount_P1 + 12

					PF_TechFrameText_P1:SetTextColor(1, 1, 1, 1)
					PF_TechFrameText_P1:SetText(PF_TECH_BONUS_TEXT)
					PF_TechFrame_P1:SetPoint("BOTTOM", PF_CellContainer_P1, "BOTTOM", 0, 16)
					PF_TechFrame_P1:SetAlpha(1.0)
					PF_TechFrame_P1:SetScript("OnUpdate", PF_TechFrameOnUpdate)
				end

				PF_AttackCount_P1 = math.ceil(PF_AttackCount_P1)

				-- Attacking can be used to counter incoming counter gems at half the rate
				if PF_AttackCount_P1 > 0 and PF_AttackedCount_P1 > 0 then
					if PF_AttackedCount_P1 * 2 >= PF_AttackCount_P1 then
						-- All attack gems used to counter incoming attacks with no remainder
						PF_AttackedCount_P1 = PF_AttackedCount_P1 - math.floor(PF_AttackCount_P1 / 2)
						PF_AttackCount_P1 = 0
					else
						-- Some attack gems used to counter all incoming attacks with remainder
						PF_AttackCount_P1 = PF_AttackCount_P1 - PF_AttackedCount_P1 * 2
						PF_AttackedCount_P1 = 0
					end
				end

				-- E) Otherwise there is no gem burst/collapsing animation so drop
				--    counter gems given by the opponent, if any, and animate this
				PF_Attack(1, PF_AttackedCount_P1)
				PF_CountFrameText_P1:SetText("")
				PF_WarningFrameText_P1:SetText("")

				PF_SendMsg("ATTK "..PF_AttackCount_P1..":"..PF_AttackedCount_P1)
				PF_AttackCount_P1 = 0
				PF_AttackedCount_P1 = 0

				-- F) And attempt to spawn a new gem (where it checks for "game over")
				PF_SpawnGem()
			end
		elseif PF_GameState == 3 then
			-- D) Collapse the field
			PF_CollapseField(1)
			PF_GameState = 2
			-- Go to (A) after animation of collapse is done (via this OnUpdate function)
		end

	elseif GetTime() - PF_Time < 0 then
		-- In case it rolls over. It shouldn't though.
		PF_Time = GetTime()
	end
end

local info = {}
function PF_CharDropDown_Initialize()
	for i = 1, #PF_CharList, 1 do
		info.text = PF_CharList[i]
		info.func = PF_CharDropDown_OnClick
		info.checked = nil
		UIDropDownMenu_AddButton(info)
	end
end

function PF_ShowPattern(player, selected)
	local selectedhero = PF_CharList[selected]

	getglobal("PF_CharPatternFrame_P"..player):Show()
	for i = 1, 24 do
		getglobal("PF_CharPattern"..i.."_P"..player):SetNormalTexture(PF_GemTextures[PF_GemPatterns[selectedhero][i] - 26])
	end
end

--Initial locations of these 2 frames
--PF_CharPatternFrame_P1:SetPoint("TOP", PF_Frame, "TOP", -146, -290)
--PF_CharPatternFrame_P2:SetPoint("TOP", PF_Frame, "TOP", 146, -290)

function PF_CharPatternFrameOnUpdate(this, elapsed)
	-- Throtle to 40 fps
	this.timer = (this.timer or 0) + elapsed
	if this.timer < 0.025 then return end
	this.timer = 0

	local t = this.anim
	t = t + 0.05

	PF_CharPatternFrame_P1:SetScale(1 - (1 - 0.39) * sqrt(t))
	PF_CharPatternFrame_P1:SetPoint("TOP", PF_Frame, "TOP", (-146+146*t), (-290-((800-290)*t)))
	if t >= 1 then
		PF_CharPatternFrame_P1:SetScript("OnUpdate", nil)
	end
	this.anim = t
end

function PF_CharPatternFrameOnUpdate2(this, elapsed)
	-- Throtle to 40 fps
	this.timer = (this.timer or 0) + elapsed
	if this.timer < 0.025 then return end
	this.timer = 0

	local t = this.anim
	t = t + 0.05

	PF_CharPatternFrame_P2:SetScale(1 - (1 - 0.39) * sqrt(t))
	PF_CharPatternFrame_P2:SetPoint("TOP", PF_Frame, "TOP", (146-146*t), (-290-((1000-290)*t)))
	if t >= 1 then
		PF_CharPatternFrame_P2:SetScript("OnUpdate", nil)
	end
	this.anim = t
end

function PF_RemoveRow15(player)
	for i = 1, 6 do
		a = PF_ID[player][i][15]
		if a > 0 then
			-- There is a button, so fade it out
			a = PF_Button[player][a]
			a:SetScript("OnUpdate", PF_FadeCell)
		end
		-- And clear the internal fields
		PF_ID[player][i][15] = 0
		PF_Field[player][i][15] = 0
	end
end


------------- Communication Functions -----------------

function PF_Challenge(opp)
	if opp == nil or opp == "" then
		return
	end

	-- Parse the name to look prettier (caps 1st char, non-caps remainder)
	opp = strupper(strsub(opp, 1, 1))..strlower(strsub(opp, 2))

	-- Check that opponent specified is not yourself
	if opp == UnitName("player") then
		local dialog = StaticPopup_Show("PF_CHALLENGE_ERROR3")
		return
	end

	if PF_OpponentName == PF_NO_OPPONENT_TEXT and PF_ChallengedOpp == PF_NO_OPPONENT_TEXT then
		-- Send the challenge!
		PF_StatusLabel:SetText(PF_STATUS_TEXT..format(PF_CHALLENGING_TEXT, opp))
		PF_ChallengedOpp = opp
		SendAddonMessage("PuzzleFighter", "CHLG challenge", "WHISPER", opp)
	else
		-- Can't challenge, already in an existing game
		local dialog = StaticPopup_Show("PF_CHALLENGE_ERROR")
	end
end

function PF_ChallengeAcceptButton()
	SendAddonMessage("PuzzleFighter", "CHLG accept", "WHISPER", PF_ChallengedOpp)
	PF_OpponentName = PF_ChallengedOpp
	-- Set up new game as client
	if not PF_Frame:IsShown() then
		if InCombatLockdown() then
			ChatFrame1:AddMessage(PF_INCOMBATLOCKDOWN)
		else
			PF_Frame:Show()
		end
	end
	PF_IsServer = false
	PlaySoundFile(PF_SoundFiles["ACCEPT"])
	PF_NewGame()
end

function PF_ChallengeDeclineButton()
	SendAddonMessage("PuzzleFighter", "CHLG decline", "WHISPER", PF_ChallengedOpp)
	PF_ChallengedOpp = PF_NO_OPPONENT_TEXT
end

function PF_InitMessageBuffer()
	for k in pairs(PF_MessageBuffer) do
		PF_MessageBuffer[k] = nil
	end
	PF_MessageBufferStart = 1
	PF_MessageBufferEnd = 1
end

function PF_StoreMessage(msg)
	PF_MessageBufferEnd = PF_MessageBufferEnd + 1
	if PF_MessageBufferEnd > PF_MESSAGE_BUFFER_MAX then PF_MessageBufferEnd = 1 end

	PF_MessageBuffer[PF_MessageBufferEnd] = msg
end

function PF_ClearMessageBuffer()
	while PF_MessageBufferStart ~= PF_MessageBufferEnd do
		PF_ProcessMessage()
	end
end

function PF_ProcessMessage()
	local msg
	local cmd, args
	local a, b, c, d

	-- Check if buffer is empty
	if PF_MessageBufferStart == PF_MessageBufferEnd then return end

	PF_MessageBufferStart = PF_MessageBufferStart + 1
	if PF_MessageBufferStart > PF_MESSAGE_BUFFER_MAX then PF_MessageBufferStart = 1 end

	msg = PF_MessageBuffer[PF_MessageBufferStart]

	cmd = strsub(msg, 1, 4)
	args = strsub(msg, 6)

	if cmd == "CHAR" and PF_InitState == 1 then
		_, _, a = strfind(args, "(%d+)")
		if a then
			PF_Hero_P2 = tonumber(a)
			UIDropDownMenu_SetSelectedID(PF_CharDropDown_P2, PF_Hero_P2)
			PF_ShowPattern(2, PF_Hero_P2)
		end
	end

	if cmd == "RCHR" and PF_Hero_P2 == #PF_CharList then
		_, _, a = strfind(args, "(%d+)")
		if a then
			PF_Hero_P2 = tonumber(a)
			UIDropDownMenu_SetSelectedID(PF_CharDropDown_P2, PF_Hero_P2)
			PF_ShowPattern(2, PF_Hero_P2)
			PF_CharPatternFrameLabel_P2:SetText(PF_CharList[PF_Hero_P2].." -->")
		end
	end

	if cmd == "REDY" and PF_InitState == 1 then
		_, _, a = strfind(args, "(%d+)")
		if a then
			PF_Hero_P2 = tonumber(a)
			UIDropDownMenu_SetSelectedID(PF_CharDropDown_P2, PF_Hero_P2)
			PF_ShowPattern(2, PF_Hero_P2)
			PF_StatusLabel:SetText(PF_STATUS_TEXT..format(PF_READY_AND_WAITING_TEXT, PF_ChallengedOpp))
			if PF_ReadyButton:IsEnabled() == 0 then
				PF_SendMsg("STRT")
				PF_StartGame()
			end
			PF_CharPatternFrame_P2:SetScript("OnUpdate", PF_CharPatternFrameOnUpdate2)
			PF_CharPatternFrameLabel_P2:ClearAllPoints()
			PF_CharPatternFrameLabel_P2:SetPoint("BOTTOMRIGHT", PF_CharPatternFrame_P2, "TOPRIGHT", -5, 2)
			PF_CharPatternFrameLabel_P2:SetText(PF_CharList[PF_Hero_P2].." -->")
			PF_CharDropDown_P2:Hide()
		end
	end

	if cmd == "STRT" and PF_InitState == 1 then
		PF_StartGame()
	end

	if cmd == "GSTR" then
		if PF_InitState == 1 and PF_IsServer == false then
			PF_GemString = args
			PF_CharDropDown_P1:Show()
			PF_CharDropDown_P2:Show()
			PF_CharPatternFrameLabel_P1:Show()
			PF_CharPatternFrameLabel_P2:Show()
			PF_ReadyButton:Enable()
			PF_ReadyButton:Show()
			PF_PauseButton:Disable()
			PF_SendMsg("CHAR "..PF_Hero_P1)
		elseif PF_InitState == 2 and PF_IsServer == false then
			PF_GemString = PF_GemString..args
		end
	end

	if cmd == "GREQ" then
		if PF_InitState == 2 and PF_IsServer == true then
			local a = PF_GenerateGemString()
			PF_GemString = PF_GemString..a
			PF_SendMsg("GSTR "..a)
		end
	end

	if cmd == "PAUS" then
		if args == "pause" then
			PF_PauseGame(true)
		elseif args == "unpause" then
			PF_UnpauseGame(true)
		end
	end

	if cmd == "SPWN" and PF_InitState == 2 then
		_, _, a, b, c, d = strfind(args, "(%d+):(%d+):(%d+):(%d+)")
		if a and b and c and d then
			a = tonumber(a)
			b = tonumber(b)
			c = tonumber(c)
			d = tonumber(d)
			if a > 0 and b > 0 and c > 0 and d > 0 then
			if a < 10 and b < 10 and c < 10 and d < 10 then
				PF_SpawnGem2(a, b)
				PF_UpdateNextGem(2, c, d)
			end
			end
		else
			return
		end
	end

	if (cmd == "MOVE" or cmd == "DROP") and PF_InitState == 2 then
		_, _, a, b, c, d = strfind(args, "(%d+):(%d+):(%d+):(%d+)")
		if a and b and c and d then
			a = tonumber(a)
			b = tonumber(b)
			c = tonumber(c)
			d = tonumber(d)
			if a > 0 and a < 7 and c > 0 and c < 7 then
			if b > 0 and b < 16 and d > 0 and d < 16 then
				PF_FallingButton_P2.TargetX = a * 32 - 22
				PF_FallingButton_P2.TargetY = b * 32 - 22
				PF_FallingButton2_P2.TargetX = c * 32 - 22
				PF_FallingButton2_P2.TargetY = d * 32 - 22
				PF_FallingButton_P2:SetScript("OnUpdate", PF_ButtonMove)
				PF_FallingButton2_P2:SetScript("OnUpdate", PF_ButtonMove)
			end
			end
		else
			return
		end
	end

	if cmd == "DROP" and PF_InitState == 2 then
		-- Store the gem locations in the internal field
		PF_Field[2][a][b] = PF_FallingButton_P2.gem
		PF_Field[2][c][d] = PF_FallingButton2_P2.gem
		PF_ID[2][a][b] = PF_FallingButton_P2:GetID()
		PF_ID[2][c][d] = PF_FallingButton2_P2:GetID()
		PF_Time2 = GetTime()
		PF_Chain_P2 = 0
		PF_CellContainer_P2:SetScript("OnUpdate", PF_OnUpdate2)
		PF_Countdown_P2 = true
		PF_Frame:SetScript("OnUpdate", nil)
	end

	if cmd == "ATTK" and PF_InitState == 2 then
		_, _, a, b = strfind(args, "(%d+):(%d+)")
		if a and b then
			a = tonumber(a)
			b = tonumber(b)
			PF_AttackedCount_P1 = PF_AttackedCount_P1 + a
			PF_AttackedCount_P2 = b
			if PF_AttackedCount_P1 > 0 then
				if PF_AttackedCount_P1 < 16 then
					PF_CountFrameText_P1:SetTextColor(1, 1, 1, 1)
					PF_WarningFrameText_P1:SetTextColor(1, 1, 1, 1)
					PF_WarningFrameText_P1:SetText(PF_CAUTION_TEXT)
				elseif PF_AttackedCount_P1 < 31 then
					PF_CountFrameText_P1:SetTextColor(1, 1, 0, 1)
					PF_WarningFrameText_P1:SetTextColor(1, 1, 0, 1)
					PF_WarningFrameText_P1:SetText(PF_WARNING_TEXT)
				else
					PF_CountFrameText_P1:SetTextColor(1, 0.5, 0, 1)
					PF_WarningFrameText_P1:SetTextColor(1, 0.5, 0, 1)
					PF_WarningFrameText_P1:SetText(PF_DANGER_TEXT)
				end
				PF_CountFrameText_P1:SetText(PF_AttackedCount_P1)
			end
			PF_CountFrameText_P2:SetText("")
			PF_WarningFrameText_P2:SetText("")
			PF_SendMsg("ATK2 "..PF_AttackedCount_P1)
		else
			return
		end
	end

	if cmd == "ATK2" and PF_InitState == 2 then
		_, _, a = strfind(args, "(%d+)")
		if a then
			a = tonumber(a)
			if a > 0 then
				if a < 16 then
					PF_CountFrameText_P2:SetTextColor(1, 1, 1, 1)
					PF_WarningFrameText_P2:SetTextColor(1, 1, 1, 1)
					PF_WarningFrameText_P2:SetText(PF_CAUTION_TEXT)
				elseif a < 31 then
					PF_CountFrameText_P2:SetTextColor(1, 1, 0, 1)
					PF_WarningFrameText_P2:SetTextColor(1, 1, 0, 1)
					PF_WarningFrameText_P2:SetText(PF_WARNING_TEXT)
				else
					PF_CountFrameText_P2:SetTextColor(1, 0.5, 0, 1)
					PF_WarningFrameText_P2:SetTextColor(1, 0.5, 0, 1)
					PF_WarningFrameText_P2:SetText(PF_DANGER_TEXT)
				end
				PF_CountFrameText_P2:SetText(a)
			end
		else
			return
		end
	end

	if cmd == "GOVR" and PF_InitState == 2 then
		PlaySoundFile(PF_SoundFiles["WIN"])

		PF_CellContainer_P1:SetScript("OnUpdate", nil)
		PF_InitState = 1
		PF_StatusLabel:SetText(PF_STATUS_TEXT..PF_GAME_OVER_WIN_TEXT)
		PF_PauseButton:SetText(START)
		PF_GameState = 0

		PF_ChainText_P1:SetTextColor(0, 1, 0, 1)
		PF_ChainText_P1:SetText(PF_YOU_WIN_TEXT)
		PF_ChainFrame_P1:SetPoint("CENTER", PF_CellContainer_P1, "CENTER", 0, -24)
		PF_ChainFrame_P1:SetAlpha(1.0)
		PF_ChainFrame_P1:SetScript("OnUpdate", PF_ChainFrameOnUpdate)

		PF_ChainText_P2:SetTextColor(1, 0, 0, 1)
		PF_ChainText_P2:SetText(PF_LOSER_TEXT)
		PF_ChainFrame_P2:SetPoint("CENTER", PF_CellContainer_P2, "CENTER", 0, -24)
		PF_ChainFrame_P2:SetAlpha(1.0)
		PF_ChainFrame_P2:SetScript("OnUpdate", PF_ChainFrameOnUpdate)

		PF_ClearMessageBuffer()
		PF_Frame:SetScript("OnUpdate", nil)

		PF_OpponentName = PF_NO_OPPONENT_TEXT
		PF_ChallengedOpp = PF_NO_OPPONENT_TEXT
	end

end


-- Function that handles all incoming communication
function PF_OnEvent(event)
	if event == "CHAT_MSG_ADDON" and arg1 == "PuzzleFighter" then
		-- Process the command
		local cmd = strsub(arg2, 1, 4)
		local args = strsub(arg2, 6)
		local a, b, c, d

		if cmd == "CHLG" then
			if args == "challenge" then
				if PF_OpponentName == PF_NO_OPPONENT_TEXT and PF_ChallengedOpp == PF_NO_OPPONENT_TEXT then
					local dialog = StaticPopup_Show("PF_CHALLENGED", arg4)
					PF_ChallengedOpp = arg4
				else
					SendAddonMessage("PuzzleFighter", "CHLG ingame", "WHISPER", arg4)
				end
			elseif args == "accept" then
				if arg4 == PF_ChallengedOpp then
					PF_StatusLabel:SetText(PF_STATUS_TEXT..format(PF_CHALLENGE_ACCEPTED_TEXT, PF_ChallengedOpp))
					PF_OpponentName = PF_ChallengedOpp
					-- Set up new game as server
					if not PF_Frame:IsShown() then
						if InCombatLockdown() then
							ChatFrame1:AddMessage(PF_INCOMBATLOCKDOWN)
						else
							PF_Frame:Show()
						end
					end
					PF_IsServer = true
					PlaySoundFile(PF_SoundFiles["ACCEPT"])
					PF_NewGame()
					PF_CharDropDown_P1:Show()
					PF_CharDropDown_P2:Show()
					PF_CharPatternFrameLabel_P1:Show()
					PF_CharPatternFrameLabel_P2:Show()
					PF_ReadyButton:Enable()
					PF_ReadyButton:Show()
					PF_PauseButton:Disable()

					PF_GemString = PF_GenerateGemString()
					PF_SendMsg("GSTR "..PF_GemString)
					PF_SendMsg("CHAR "..PF_Hero_P1)
				end
			elseif args == "decline" then
				if arg4 == PF_ChallengedOpp then
					PF_StatusLabel:SetText(PF_STATUS_TEXT..PF_READY_TEXT.." - "..format(PF_CHALLENGE_DECLINED_TEXT, PF_ChallengedOpp))
					local dialog = StaticPopup_Show("PF_DECLINE", PF_ChallengedOpp)
					PF_ChallengedOpp = PF_NO_OPPONENT_TEXT
				end
			elseif args == "ingame" then
				if arg4 == PF_ChallengedOpp then
					PF_StatusLabel:SetText(PF_STATUS_TEXT..PF_READY_TEXT.." - "..format(PF_ALREADY_IN_GAME_TEXT, PF_ChallengedOpp))
					local dialog = StaticPopup_Show("PF_CHALLENGE_ERROR2", PF_ChallengedOpp)
					PF_ChallengedOpp = PF_NO_OPPONENT_TEXT
				end
			end
		end

		-- Process the remainder commands only if the sender is the opponent
		if arg4 == PF_OpponentName then
			-- Add them to a message buffer
			PF_StoreMessage(arg2)
		end

	elseif event == "VARIABLES_LOADED" then
		if PF_Options                 == nil then PF_Options                 = {}      end
		if PF_Options["LockWindow"]   == nil then PF_Options["LockWindow"]   = false   end
		if PF_Options["DownButton"]   == nil then PF_Options["DownButton"]   = true    end
		if PF_Options["SoundEffects"] == nil then PF_Options["SoundEffects"] = true    end
		if PF_Options["Scaling"]      == nil then PF_Options["Scaling"]      = 100     end
		if PF_Options["Transparency"] == nil then PF_Options["Transparency"] = 100     end
		if PF_Options["Left"]         == nil then PF_Options["Left"]         = "LEFT"  end
		if PF_Options["Right"]        == nil then PF_Options["Right"]        = "RIGHT" end
		if PF_Options["Down"]         == nil then PF_Options["Down"]         = "DOWN"  end
		if PF_Options["Rotate"]       == nil then PF_Options["Rotate"]       = "UP"    end

		PF_SetSoundEffects()
		PF_Frame:SetScale(PF_Options["Scaling"] / 100)
		PF_Frame:SetAlpha(PF_Options["Transparency"] / 100)

	elseif event == "PLAYER_REGEN_DISABLED" then
		-- This causes an existing game to pause and revert the override bindings
		-- to the default stuff.
		PF_Frame:Hide()

	end
end

StaticPopupDialogs["PF_CHALLENGE"] = {
	text = PF_CHALLENGE_PROMPT,
	button1 = PF_CHALLENGE2_TEXT,
	button2 = CANCEL,
	hasEditBox = 1,
	maxLetters = 12,
	OnAccept = function()
		local editBox = getglobal(this:GetParent():GetName().."EditBox")
		PF_Challenge(editBox:GetText())
	end,
	OnShow = function()
		getglobal(this:GetName().."EditBox"):SetFocus()
	end,
	OnHide = function()
		if ChatFrameEditBox:IsVisible() then
			ChatFrameEditBox:SetFocus()
		end
		getglobal(this:GetName().."EditBox"):SetText("")
	end,
	EditBoxOnEnterPressed = function()
		local editBox = getglobal(this:GetParent():GetName().."EditBox")
		PF_Challenge(editBox:GetText())
		this:GetParent():Hide()
	end,
	EditBoxOnEscapePressed = function()
		this:GetParent():Hide()
	end,
	timeout = 0,
	exclusive = 1,
	whileDead = 1,
	hideOnEscape = 1
}

StaticPopupDialogs["PF_CHALLENGED"] = {
	text = PF_CHALLENGE_PROMPT2,
	button1 = YES,
	button2 = NO,
	OnAccept = function()
		PF_ChallengeAcceptButton()
	end,
	OnCancel = function()
		PF_ChallengeDeclineButton()
	end,
	sound = "igPlayerInvite",
	timeout = 0,
	whileDead = 1,
	hideOnEscape = 1
}

StaticPopupDialogs["PF_CHALLENGE_ERROR"] = {
	text = PF_CHALLENGE_ERROR,
	button1 = TEXT(OKAY),
	showAlert = 1,
	timeout = 0,
}

StaticPopupDialogs["PF_CHALLENGE_ERROR2"] = {
	text = PF_CHALLENGE_ERROR2,
	button1 = TEXT(OKAY),
	showAlert = 1,
	timeout = 0,
}

StaticPopupDialogs["PF_CHALLENGE_ERROR3"] = {
	text = PF_CHALLENGE_ERROR3,
	button1 = TEXT(OKAY),
	showAlert = 1,
	timeout = 0,
}

StaticPopupDialogs["PF_DECLINE"] = {
	text = PF_CHALLENGE_DECLINE,
	button1 = TEXT(OKAY),
	showAlert = 1,
	timeout = 0,
}

---------------- Functions specific to the 2nd player -----------------

-- Function to display the spawning gem of the opponent field
-- given incoming information
function PF_SpawnGem2(color1, color2)
	-- Get an unused button and update that button on screen for Gem 1
	PF_FallingButton_P2 = PF_GetUnusedButton(2)
	PF_FallingButton_P2.gem = color1
	PF_FallingButton_P2.SizeX = 1
	PF_FallingButton_P2.SizeY = 1
	PF_FallingButton_P2:SetWidth(32)
	PF_FallingButton_P2:SetHeight(32)
	PF_DrawCell(PF_FallingButton_P2, 4, 14)

	-- Get an unused button and update that button on screen for Gem 2
	PF_FallingButton2_P2 = PF_GetUnusedButton(2)
	PF_FallingButton2_P2.gem = color2
	PF_FallingButton2_P2.SizeX = 1
	PF_FallingButton2_P2.SizeY = 1
	PF_FallingButton2_P2:SetWidth(32)
	PF_FallingButton2_P2:SetHeight(32)
	PF_DrawCell(PF_FallingButton2_P2, 4, 15)
end

local PF_GameState2 = true
-- Function to delay/stagger the animation of chain crash/collapsing gems
-- of the opponent playing field
function PF_OnUpdate2()
	if GetTime() - PF_Time2 >= PF_ANIMTIME then
		PF_Time2 = PF_Time2 + PF_ANIMTIME

		PF_RemoveRow15(2)

		if PF_Countdown_P2 == true then
			PF_CountdownCounterGems(2)
			PF_Countdown_P2 = false
		end

		while PF_CheckCoalesce(2) == true do end
		if PF_GameState2 and PF_CheckGemBurst(2) > 0 then
			PF_PerformGemBurst(2)
			PF_GameState2 = false

			-- Display chains above 1
			PF_Chain_P2 = PF_Chain_P2 + 1
			if PF_Chain_P2 > 1 then
				PF_ChainText_P2:SetTextColor(1, 0, 0, 1)
				PF_ChainText_P2:SetText(PF_Chain_P2..PF_CHAIN_TEXT)
				PF_ChainFrame_P2:SetPoint("CENTER", PF_CellContainer_P2, "CENTER", 0, -24)
				PF_ChainFrame_P2:SetAlpha(1.0)
				PF_ChainFrame_P2:SetScript("OnUpdate", PF_ChainFrameOnUpdate)
			end
		elseif not PF_GameState2 then
			PF_CollapseField(2)
			PF_GameState2 = true
		else
			PF_ProcessMessage()
			if PF_AttackedCount_P2 >= 0 then
				-- Check for empty field and display a tech bonus message
				if PF_CheckEmptyField(2) == true then
					PF_TechFrameText_P2:SetTextColor(1, 1, 1, 1)
					PF_TechFrameText_P2:SetText(PF_TECH_BONUS_TEXT)
					PF_TechFrame_P2:SetPoint("BOTTOM", PF_CellContainer_P2, "BOTTOM", 0, 16)
					PF_TechFrame_P2:SetAlpha(1.0)
					PF_TechFrame_P2:SetScript("OnUpdate", PF_TechFrameOnUpdate)
				end

				PF_Attack(2, PF_AttackedCount_P2)
				PF_AttackedCount_P2 = -1
				PF_CellContainer_P2:SetScript("OnUpdate", nil)
				PF_Frame:SetScript("OnUpdate", PF_ProcessMessage)
			end
		end

	elseif GetTime() - PF_Time2 < 0 then
		-- In case it rolls over midnight hour.
		PF_Time2 = GetTime()
	end
end


--------- Game Functions -----------

function PF_OnLoad()
	local a, b, c, d, e

	-- Hide the frame when escape is pressed
	tinsert(UISpecialFrames, "PF_Frame")

	-- Register the slash command
	SLASH_PUZZLEFIGHTER1 = "/pf"
	SLASH_PUZZLEFIGHTER2 = "/puzzlefighter"
	SlashCmdList["PUZZLEFIGHTER"] = function(msg)
		PF_SlashCommand(msg)
	end

	this:RegisterEvent("CHAT_MSG_ADDON")
	this:RegisterEvent("VARIABLES_LOADED")
	this:RegisterEvent("PLAYER_REGEN_DISABLED")

	for i = 1, 86 do
		a = getglobal("PF_Cell"..i.."_P1")
		a:SetBackdropColor(0, 0, 0, 0)	-- Remove the backdrop color
		a.container = PF_CellContainer_P1
		PF_Button[1][i] = a		-- Assign to a button array

		b = getglobal("PF_Cell"..i.."_P1Label")
		c, d, e = b:GetFont()		-- Change the font size to 26
		b:SetFont(c, 26, e)

		a.border = getglobal("PF_Cell"..i.."_P1Border")
		a.label = b

		-- Repeat for player 2
		a = getglobal("PF_Cell"..i.."_P2")
		a:SetBackdropColor(0, 0, 0, 0)
		a.container = PF_CellContainer_P2
		PF_Button[2][i] = a

		b = getglobal("PF_Cell"..i.."_P2Label")
		c, d, e = b:GetFont()
		b:SetFont(c, 26, e)

		a.border = getglobal("PF_Cell"..i.."_P2Border")
		a.label = b
	end

	PF_CharDropDown_P2Button:Disable()
	PF_ReadyButton:Hide()

	-- This PF_Reset() must come after the button array assignment above
	PF_Reset()
end

function PF_SlashCommand(msg)
	PF_Toggle()
end

function PF_Toggle()
	if PF_Frame:IsShown() then
		PF_Frame:Hide()
	else
		if InCombatLockdown() then
			ChatFrame1:AddMessage(PF_INCOMBATLOCKDOWN)
		else
			PF_Frame:Show()
		end
	end
end

function PF_OnHide()
	PF_PauseGame()
end

function PF_PauseToggle()
	if this:GetText() == START then
		PF_NewGame()
	elseif this:GetText() == PF_PAUSE_TEXT then
		PF_PauseGame()
	elseif this:GetText() == PF_UNPAUSE_TEXT then
		PF_UnpauseGame()
	end
end

function PF_Reset(NewGame)
	PF_InitState = 1

	if NewGame ~= true then
		-- Do not perform these when called by PF_NewGame()
		PF_OpponentName = PF_NO_OPPONENT_TEXT
		PF_ChallengedOpp = PF_NO_OPPONENT_TEXT
		PF_StatusLabel:SetText(PF_STATUS_TEXT..PF_READY_TEXT)
	end

	PF_Next1_P1:SetNormalTexture(PF_GemTextures[1])
	PF_Next2_P1:SetNormalTexture(PF_GemTextures[2])
	PF_Next1_P2:SetNormalTexture(PF_GemTextures[3])
	PF_Next2_P2:SetNormalTexture(PF_GemTextures[4])

	PF_CharDropDown_P1:Hide()
	PF_CharDropDown_P2:Hide()
	PF_CharPatternFrame_P1:Hide()
	PF_CharPatternFrame_P2:Hide()
	PF_CharPatternFrameLabel_P1:Hide()
	PF_CharPatternFrameLabel_P2:Hide()
	PF_CharPatternFrameLabel_P1:SetText(PF_COUNTER_GEM_PATTERN_TEXT)
	PF_CharPatternFrameLabel_P2:SetText(PF_COUNTER_GEM_PATTERN_TEXT)
	PF_CharPatternFrameLabel_P2:ClearAllPoints()
	PF_CharPatternFrameLabel_P2:SetPoint("BOTTOMLEFT", PF_CharPatternFrame_P2, "TOPLEFT", 5, 2)
	PF_CharPatternFrame_P1:SetScale(1.0)
	PF_CharPatternFrame_P2:SetScale(1.0)
	PF_CharPatternFrame_P1:SetPoint("TOP", PF_Frame, "TOP", -146, -290)
	PF_CharPatternFrame_P2:SetPoint("TOP", PF_Frame, "TOP", 146, -290)
	PF_CellContainer_P1:SetScript("OnUpdate", nil)
	PF_CellContainer_P2:SetScript("OnUpdate", nil)
	PF_PauseButton:SetText(START)
	PF_PauseButton:UnlockHighlight()
	PF_PauseButton:Enable()
	PF_ReadyButton:Enable()
	PF_GamePaused = false
	PF_GemString = ""
	PF_OpponentLabel:SetText(PF_OPPONENT_TEXT..PF_OpponentName)
	PF_Countdown_P1 = false
	PF_Countdown_P2 = false
	PF_AttackCount_P1 = 0
	PF_AttackedCount_P1 = 0
	PF_AttackedCount_P2 = -1
	PF_CharPatternFrame_P1.anim = 0
	PF_CharPatternFrame_P2.anim = 0
	PF_ChainText_P1:SetText("")
	PF_ChainText_P2:SetText("")
	PF_CountFrameText_P1:SetText("")
	PF_CountFrameText_P2:SetText("")
	PF_WarningFrameText_P1:SetText("")
	PF_WarningFrameText_P2:SetText("")
	PF_ReadyButton:Hide()
	PF_InitMessageBuffer()
	PF_Frame:SetScript("OnUpdate", nil)
	PF_PauseDelay = 0

	-- Clear and Hide all buttons
	for i = 1, 86 do
		PF_HideCell(PF_Button[1][i])
		PF_HideCell(PF_Button[2][i])
	end

	-- Clear the internal playing field
	for i = 1, 6 do
		for j = 1, 15 do
			PF_Field[1][i][j] = 0
			PF_ID[1][i][j] = 0
			PF_Field[2][i][j] = 0
			PF_ID[2][i][j] = 0
		end
	end

	-- For debugging purposes, to play against self!
	if IsShiftKeyDown() and IsControlKeyDown() then
		PF_OpponentName = UnitName("player")
		PF_ChallengedOpp = UnitName("player")
		PF_Hero_P2 = PF_Hero_P1
	end
end

function PF_MoveLeft()
	local player = 1

	if PF_InitState ~= 2 or PF_GamePaused == true or PF_GameState ~= 1 then
		return false
	end

	local check1, check2
	check1 = false
	check2 = false

	-- move left only if column 2 or higher, and if the space is not filled
	if PF_FallingX > 1 and PF_Field[player][PF_FallingX-1][PF_FallingY] == 0 then
		check1 = true
	end
	if PF_Falling2X > 1 and PF_Field[player][PF_Falling2X-1][PF_Falling2Y] == 0 then
		check2 = true
	end

	if check1 and check2 then
		PlaySoundFile(PF_SoundFiles["LEFT"])
		PF_FallingX = PF_FallingX - 1	
		PF_FallingButton_P1.TargetX = PF_FallingX * 32 - 22
		PF_FallingButton_P1:SetScript("OnUpdate", PF_ButtonMove)
		PF_Falling2X = PF_Falling2X - 1	
		PF_FallingButton2_P1.TargetX = PF_Falling2X * 32 - 22
		PF_FallingButton2_P1:SetScript("OnUpdate", PF_ButtonMove)
		PF_SendMsg("MOVE "..PF_FallingX..":"..PF_FallingY..":"..PF_Falling2X..":"..PF_Falling2Y)
		return true
	end
	return false
end

function PF_MoveRight()
	local player = 1

	if PF_InitState ~= 2 or PF_GamePaused == true or PF_GameState ~= 1 then
		return false
	end

	local check1, check2
	check1 = false
	check2 = false

	-- move right only if column 5 or lower, and if the space is not filled
	if PF_FallingX < 6 and PF_Field[player][PF_FallingX+1][PF_FallingY] == 0 then
		check1 = true
	end
	if PF_Falling2X < 6 and PF_Field[player][PF_Falling2X+1][PF_Falling2Y] == 0 then
		check2 = true
	end

	if check1 and check2 then
		PlaySoundFile(PF_SoundFiles["RIGHT"])
		PF_FallingX = PF_FallingX + 1	
		PF_FallingButton_P1.TargetX = PF_FallingX * 32 - 22
		PF_FallingButton_P1:SetScript("OnUpdate", PF_ButtonMove)
		PF_Falling2X = PF_Falling2X + 1	
		PF_FallingButton2_P1.TargetX = PF_Falling2X * 32 - 22
		PF_FallingButton2_P1:SetScript("OnUpdate", PF_ButtonMove)
		PF_SendMsg("MOVE "..PF_FallingX..":"..PF_FallingY..":"..PF_Falling2X..":"..PF_Falling2Y)
		return true
	end
	return false
end

function PF_MoveDown(byTimer, dropAll)
	local player = 1

	if PF_InitState ~= 2 or PF_GamePaused == true or PF_GameState ~= 1 then
		return false
	end

	local check1, check2
	check1 = false
	check2 = false

	-- move down only if row 2 or higher, and if the space is not filled
	if PF_FallingY > 1 and PF_Field[player][PF_FallingX][PF_FallingY-1] == 0 then
		check1 = true
	end
	if PF_Falling2Y > 1 and PF_Field[player][PF_Falling2X][PF_Falling2Y-1] == 0 then
		check2 = true
	end

	if check1 and check2 then
		-- Move both gems down
		if byTimer == true then
			PlaySoundFile(PF_SoundFiles["DOWNT"])
		else
			PlaySoundFile(PF_SoundFiles["DOWN"])
		end
		PF_FallingY = PF_FallingY - 1
		PF_FallingButton_P1.TargetY = PF_FallingY * 32 - 22
		PF_FallingButton_P1:SetScript("OnUpdate", PF_ButtonMove)
		PF_Falling2Y = PF_Falling2Y - 1
		PF_FallingButton2_P1.TargetY = PF_Falling2Y * 32 - 22
		PF_FallingButton2_P1:SetScript("OnUpdate", PF_ButtonMove)
		if dropAll ~= true then
			PF_SendMsg("MOVE "..PF_FallingX..":"..PF_FallingY..":"..PF_Falling2X..":"..PF_Falling2Y)
		end
		return true
	end

	-- Otherwise the gem can't move down as a pair, so break them up and drop them both
	PlaySoundFile(PF_SoundFiles["DROP"])
	for i = PF_FallingY, 2, -1 do
		if PF_Field[player][PF_FallingX][i-1] == 0 then
			PF_FallingY = PF_FallingY - 1
		else
			break
		end
	end
	if PF_GemDirection == 3 then    -- Account for the other gem below this one
		PF_FallingY = PF_FallingY + 1
	end

	for i = PF_Falling2Y, 2, -1 do
		if PF_Field[player][PF_Falling2X][i-1] == 0 then
			PF_Falling2Y = PF_Falling2Y - 1	
		else
			break
		end
	end
	if PF_GemDirection == 1 then    -- Account for the other gem below this one
		PF_Falling2Y = PF_Falling2Y + 1
	end

	PF_FallingButton_P1.TargetY = PF_FallingY * 32 - 22
	PF_FallingButton2_P1.TargetY = PF_Falling2Y * 32 - 22
	PF_FallingButton_P1:SetScript("OnUpdate", PF_ButtonMove)
	PF_FallingButton2_P1:SetScript("OnUpdate", PF_ButtonMove)

	-- Store the gem locations in the internal field
	PF_Field[player][PF_FallingX][PF_FallingY] = PF_FallingButton_P1.gem
	PF_Field[player][PF_Falling2X][PF_Falling2Y] = PF_FallingButton2_P1.gem
	PF_ID[player][PF_FallingX][PF_FallingY] = PF_FallingButton_P1:GetID()
	PF_ID[player][PF_Falling2X][PF_Falling2Y] = PF_FallingButton2_P1:GetID()

	PF_SendMsg("DROP "..PF_FallingX..":"..PF_FallingY..":"..PF_Falling2X..":"..PF_Falling2Y)

	PF_Countdown_P1 = true
	PF_AttackCount_P1 = 0
	PF_GameState = 2
	return false
end

-- Function that drops the piece down permanently
function PF_DropDown()
	-- Yes, an empty loop
	while PF_MoveDown(false, true) == true do end
end

function PF_MoveDownButton()
	if PF_Options["DownButton"] then
		PF_DropDown()
	else
		PF_MoveDown()
	end
end

function PF_Rotate()
	local player = 1

	if PF_InitState ~= 2 or PF_GamePaused == true or PF_GameState ~= 1 then
		return false
	end

	-- Check up to the next 3 rotatable positions
	for i = 1, 3 do
		PF_GemDirection = PF_GemDirection % 4 + 1

		if PF_GemDirection == 1 then
			-- Check above
			if PF_Field[player][PF_FallingX][PF_FallingY+1] == 0 then
				PlaySoundFile(PF_SoundFiles["ROTATE"])
				PF_Falling2X = PF_FallingX
				PF_Falling2Y = PF_FallingY + 1
				PF_FallingButton2_P1.TargetX = PF_Falling2X * 32 - 22
				PF_FallingButton2_P1.TargetY = PF_Falling2Y * 32 - 22
				PF_FallingButton2_P1:SetScript("OnUpdate", PF_ButtonMove)
				PF_SendMsg("MOVE "..PF_FallingX..":"..PF_FallingY..":"..PF_Falling2X..":"..PF_Falling2Y)
				return true
			end
		elseif PF_GemDirection == 2 then
			-- Check the right
			if PF_FallingX < 6 and PF_Field[player][PF_FallingX+1][PF_FallingY] == 0 then
				PlaySoundFile(PF_SoundFiles["ROTATE"])
				PF_Falling2X = PF_FallingX + 1
				PF_Falling2Y = PF_FallingY
				PF_FallingButton2_P1.TargetX = PF_Falling2X * 32 - 22
				PF_FallingButton2_P1.TargetY = PF_Falling2Y * 32 - 22
				PF_FallingButton2_P1:SetScript("OnUpdate", PF_ButtonMove)
				PF_SendMsg("MOVE "..PF_FallingX..":"..PF_FallingY..":"..PF_Falling2X..":"..PF_Falling2Y)
				return true
			end
		elseif PF_GemDirection == 3 then
			-- Check below
			if PF_FallingY > 1 and PF_Field[player][PF_FallingX][PF_FallingY-1] == 0 then
				PlaySoundFile(PF_SoundFiles["ROTATE"])
				PF_Falling2X = PF_FallingX
				PF_Falling2Y = PF_FallingY - 1
				PF_FallingButton2_P1.TargetX = PF_Falling2X * 32 - 22
				PF_FallingButton2_P1.TargetY = PF_Falling2Y * 32 - 22
				PF_FallingButton2_P1:SetScript("OnUpdate", PF_ButtonMove)
				PF_SendMsg("MOVE "..PF_FallingX..":"..PF_FallingY..":"..PF_Falling2X..":"..PF_Falling2Y)
				return true
			end
		elseif PF_GemDirection == 4 then
			-- Check the left
			if PF_FallingX > 1 and PF_Field[player][PF_FallingX-1][PF_FallingY] == 0 then
				PlaySoundFile(PF_SoundFiles["ROTATE"])
				PF_Falling2X = PF_FallingX - 1
				PF_Falling2Y = PF_FallingY
				PF_FallingButton2_P1.TargetX = PF_Falling2X * 32 - 22
				PF_FallingButton2_P1.TargetY = PF_Falling2Y * 32 - 22
				PF_FallingButton2_P1:SetScript("OnUpdate", PF_ButtonMove)
				PF_SendMsg("MOVE "..PF_FallingX..":"..PF_FallingY..":"..PF_Falling2X..":"..PF_Falling2Y)
				return true
			end
		end

		-- If the function reaches this point, it means it was
		-- unsuccesful on rotate, loop up to 3 times to check 3 directions
	end

	-- If still unsuccessful, return to original direction
	PF_GemDirection = PF_GemDirection % 4 + 1
	return false
end

function PF_CharDropDown_OnShow(player)
	local frame = getglobal("PF_CharDropDown_P"..player)

	UIDropDownMenu_Initialize(frame, PF_CharDropDown_Initialize)
	UIDropDownMenu_SetWidth(175, frame)
	UIDropDownMenu_SetSelectedID(frame, PF_Hero_P1)
	if player == 1 then
		PF_ShowPattern(1, PF_Hero_P1)
	else
		PF_ShowPattern(2, PF_Hero_P2)
	end
end

function PF_CharDropDown_OnClick()
	-- Note: Hardcorded to for P1 dropdown menu (since P2 dropdownmenu is disabled)
	local selected = this:GetID()
	UIDropDownMenu_SetSelectedID(PF_CharDropDown_P1, selected)

	PF_Hero_P1 = selected
	PF_ShowPattern(1, PF_Hero_P1)
	PF_SendMsg("CHAR "..PF_Hero_P1)
end

function PF_ReadyButton_OnClick()
	if PF_OpponentName == PF_NO_OPPONENT_TEXT or PF_OpponentName == UnitName("player") then
		PF_CharPatternFrame_P2:SetScript("OnUpdate", PF_CharPatternFrameOnUpdate2)
		PF_CharPatternFrameLabel_P2:ClearAllPoints()
		PF_CharPatternFrameLabel_P2:SetPoint("BOTTOMRIGHT", PF_CharPatternFrame_P2, "TOPRIGHT", -5, 2)
		PF_CharPatternFrameLabel_P2:SetText(PF_CharList[PF_Hero_P2].." -->")
		PF_CharDropDown_P2:Hide()
		PF_StartGame()
	else
		PF_SendMsg("REDY "..PF_Hero_P1)
		PF_ReadyButton:Disable()
		PF_StatusLabel:SetText(PF_STATUS_TEXT..format(PF_WAITING_FOR_OPP_TEXT, PF_ChallengedOpp))
	end
	PF_CharPatternFrame_P1:SetScript("OnUpdate", PF_CharPatternFrameOnUpdate)
	PF_CharPatternFrameLabel_P1:SetText("<-- "..PF_CharList[PF_Hero_P1])
	PF_CharDropDown_P1:Hide()
end

--------- Puzzle Fighter Option Functions ------------

function PF_Options_OnShow()
	PF_OptionsFrame_LockWindow:SetChecked(PF_Options["LockWindow"])
	PF_OptionsFrame_DownButton:SetChecked(PF_Options["DownButton"])
	PF_OptionsFrame_SoundEffects:SetChecked(PF_Options["SoundEffects"])
	PF_ScaleSliderText:SetText(PF_SCALING_TEXT.." "..PF_ScaleSlider:GetValue().."%")
	PF_ScaleSlider:SetValue(PF_Options["Scaling"])
	PF_TransparencySliderText:SetText(PF_TRANSPARENCY_TEXT.." "..PF_TransparencySlider:GetValue().."%")
	PF_TransparencySlider:SetValue(PF_Options["Transparency"])
end

function PF_Options_Done()
	PF_OptionsFrame:Hide()
end

function PF_Options_LockWindowToggle()
	PF_Options["LockWindow"] = not PF_Options["LockWindow"]
	PF_OptionsFrame_LockWindow:SetChecked(PF_Options["LockWindow"])
end

function PF_Options_DownButtonToggle()
	PF_Options["DownButton"] = not PF_Options["DownButton"]
	PF_OptionsFrame_DownButton:SetChecked(PF_Options["DownButton"])
end

function PF_Options_SoundEffectsToggle()
	PF_Options["SoundEffects"] = not PF_Options["SoundEffects"]
	PF_SetSoundEffects()
	PF_OptionsFrame_SoundEffects:SetChecked(PF_Options["SoundEffects"])
end

function PF_SetSoundEffects()
	if PF_Options["SoundEffects"] then
		PF_SoundFiles = PF_SOUND_FILE_DEFAULTS
	else
		PF_SoundFiles = PF_NO_SOUND_FILE_DEFAULTS
	end
end

--------- Debug Functions ------------
--[[
function PF_SpamID(player, a)
	local b
	for j = a, 1, -1 do
		b = ""
		for i = 1, 6 do
			b = b..PF_ID[player][i][j]..","
		end
		ChatFrame1:AddMessage(b)
	end
end

function PF_SpamColor(player, a)
	local b
	for j = a, 1, -1 do
		b = ""
		for i = 1, 6 do
			b = b..PF_Field[player][i][j]..","
		end
		ChatFrame1:AddMessage(b)
	end
end
]]
